import json
import os
from typing import Dict
from lpsim.agents.interaction_agent import InteractionAgent_V1_0, InteractionAgent
from lpsim.agents.nothing_agent import NothingAgent
from lpsim.server.match import Match, MatchState
from lpsim.server.deck import Deck
from tests.utils_for_test import (
    check_usage,
    do_log_tests,
    get_pidx_cidx,
    get_test_id_from_command,
    remove_ids,
    set_16_omni,
    check_hp,
    make_respond,
    get_random_state,
)


def test_rana():
    """
    first: test one round one time, can use imeediately, only elemental skill
    will trigger, can trigger multiple Rana.
    """
    agent_0 = NothingAgent(player_idx=0)
    agent_1 = InteractionAgent_V1_0(
        version="1.0",
        player_idx=1,
        verbose_level=0,
        commands=[
            "sw_card",
            "choose 0",
            "reroll 1 2 3 4 5 6 7",
            "card 0 0 omni geo",
            "skill 1 dendro dendro omni",
            "card 0 0 hydro hydro",
            "tune pyro 0",
            "card 0 0 dendro dendro",
            "end",
            "reroll 0 1 2 3 4 5 6 7",
            "skill 0 dendro pyro anemo",
            "tune hydro 0",
            "tune hydro 0",
            "skill 2 dendro dendro omni",
            "end",
            "reroll 3 4 5 6 7",
            "skill 1 dendro dendro dendro",
            "card 0 0 dendro dendro",
            "skill 1 dendro dendro dendro",
            "skill 0 dendro geo anemo",
        ],
        only_use_command=True,
    )
    match = Match(version="0.0.1", random_state=get_random_state())
    deck = {
        "name": "Deck",
        "characters": [
            {
                "name": "DendroMobMage",
                "element": "DENDRO",
                "hp": 99,
                "max_hp": 99,
            },
            {
                "name": "DendroMobMage",
                "element": "DENDRO",
            },
            {
                "name": "ElectroMobMage",
                "element": "ELECTRO",
            },
        ],
        "cards": [
            {
                "name": "Rana",
            }
        ]
        * 30,
    }
    deck = Deck(**deck)
    match.set_deck([deck, deck])
    match.config.max_same_card_number = 30
    assert match.start()[0]
    match.step()

    while True:
        if match.need_respond(0):
            make_respond(agent_0, match)
        elif match.need_respond(1):
            if len(agent_1.commands) == 6:
                # should only have two electro dice
                assert len(match.player_tables[1].dice.colors) == 2
                assert match.player_tables[1].dice.colors == ["ELECTRO", "ELECTRO"]
                assert match.player_tables[1].dice.colors[0] == "ELECTRO"
                assert match.player_tables[1].dice.colors[1] == "ELECTRO"
            make_respond(agent_1, match)
        else:
            raise AssertionError("No need respond.")
        if len(agent_1.commands) == 0:
            break

    assert len(agent_1.commands) == 0
    assert match.round_number == 3
    check_hp(match, [[83, 10, 10], [99, 10, 10]])
    assert len(match.player_tables[1].dice.colors) == 1
    assert match.player_tables[1].dice.colors[0] == "PYRO"
    assert len(match.player_tables[1].supports) == 4
    for support in match.player_tables[1].supports:
        assert support.name == "Rana"

    assert match.state != MatchState.ERROR

    """
    second: next one is other people; cannot generate when only one character;
    TODO: if overcharged self, will generate next of next?
    """
    agent_0 = InteractionAgent_V1_0(
        version="1.0",
        player_idx=0,
        verbose_level=0,
        commands=[
            "sw_card",
            "choose 2",
            "skill 0 omni omni omni",
            "end",
            "skill 0 omni omni omni",
            "end",
        ],
        only_use_command=True,
    )
    agent_1 = InteractionAgent_V1_0(
        version="1.0",
        player_idx=1,
        verbose_level=0,
        commands=[
            "sw_card",
            "choose 0",
            "choose 2",
            "card 0 0 omni omni",
            "card 1 0 omni omni",
            "skill 1 omni omni omni",
            "sw_char 1 omni",
            "end",
            "choose 2",
            "skill 1 omni omni omni",
            "end",
        ],
        only_use_command=True,
    )
    match = Match(version="0.0.1", random_state=get_random_state())
    deck = {
        "name": "Deck",
        "characters": [
            {
                "name": "PyroMobMage",
                "element": "PYRO",
                "hp": 1,
                "max_hp": 1,
            },
            {
                "name": "DendroMobMage",
                "element": "DENDRO",
                "hp": 1,
                "max_hp": 1,
            },
            {
                "name": "ElectroMobMage",
                "element": "ELECTRO",
            },
        ],
        "cards": [
            {
                "name": "Rana",
            }
        ]
        * 30,
    }
    deck = Deck(**deck)
    match.set_deck([deck, deck])
    match.config.max_same_card_number = 30
    match.config.random_first_player = False
    set_16_omni(match)
    assert match.start()[0]
    match.step()

    while True:
        if match.need_respond(0):
            make_respond(agent_0, match)
        elif match.need_respond(1):
            if len(agent_1.commands) == 4:
                # should be 8 omni and 2 dendro
                assert len(match.player_tables[1].dice.colors) == 10
                omni_num = 8
                dendro_num = 2
                for die in match.player_tables[1].dice.colors:
                    if die == "OMNI":
                        omni_num -= 1
                    else:
                        assert die == "DENDRO"
                        dendro_num -= 1
                assert omni_num == 0
                assert dendro_num == 0
            elif len(agent_1.commands) == 1:
                # should be 13 omni
                assert len(match.player_tables[1].dice.colors) == 13
                for die in match.player_tables[1].dice.colors:
                    assert die == "OMNI"
            make_respond(agent_1, match)
        else:
            raise AssertionError("No need respond.")
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    assert len(agent_1.commands) == 0
    assert len(agent_0.commands) == 0
    assert match.round_number == 3
    check_hp(match, [[1, 1, 4], [0, 0, 10]])
    assert len(match.player_tables[1].supports) == 2
    for support in match.player_tables[1].supports:
        assert support.name == "Rana"

    assert match.state != MatchState.ERROR


def test_timmie():
    """ """
    agent_0 = InteractionAgent(
        player_idx=0,
        verbose_level=0,
        commands=[
            "sw_card",
            "choose 0",
            "reroll",
            "card 0 0",
            "end",
            "reroll",
            "card 0 0",
            "card 4 0 6 7",
            "end",
            "TEST 11 record current dice color",
            "reroll",
            "TEST 1 all have 1 omni, compare with previous.0 tm2 1 tm2*3",
            "end",
            "reroll",
        ],
        only_use_command=True,
    )
    agent_1 = InteractionAgent(
        player_idx=1,
        verbose_level=0,
        commands=[
            "sw_card",
            "choose 0",
            "reroll",
            "card 0 0",
            "card 2 0",
            "end",
            "reroll",
            "card 1 0 1",
            "card 1 0",
            "card 2 0",
            "card 2 1",
            "end",
            "reroll",
            "end",
            "TEST 22 record current dice color",
            "reroll",
            "TEST 2 check hand card number",
            "end",
        ],
        only_use_command=True,
    )
    match = Match(random_state=get_random_state())
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Fischl
        character:Mona
        character:Nahida
        Wine-Stained Tricorne*2
        Timmie*2
        Rana*2
        Strategize*2
        # not implement use timmie fill
        Timmie*22
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = 30
    assert match.start()[0]
    match.step()

    last_colors = [[], []]
    while True:
        if match.need_respond(0):
            while True:
                test_id = get_test_id_from_command(agent_0)
                if test_id == 1:
                    current_colors = [
                        match.player_tables[0].dice.colors.copy(),
                        match.player_tables[1].dice.colors.copy(),
                    ]
                    assert len(current_colors[0]) == 9
                    assert len(current_colors[1]) == 9
                    assert current_colors[0][0] == "OMNI"
                    assert current_colors[1][0] == "OMNI"
                    assert current_colors[0][1:] == last_colors[0]
                    assert current_colors[1][1:] == last_colors[1]
                    tmcount = 0
                    for support in match.player_tables[0].supports:
                        if support.name == "Timmie":
                            assert support.usage == 2
                            tmcount += 1
                    assert tmcount == 1
                    tmcount = 0
                    for support in match.player_tables[1].supports:
                        if support.name == "Timmie":  # pragma: no branch
                            assert support.usage == 2
                            tmcount += 1
                    assert tmcount == 3
                elif test_id == 11:
                    last_colors = [
                        match.player_tables[0].dice.colors.copy(),
                        match.player_tables[1].dice.colors.copy(),
                    ]
                else:
                    break
            make_respond(agent_0, match)
        elif match.need_respond(1):
            while True:
                test_id = get_test_id_from_command(agent_1)
                if test_id == 2:
                    current_colors = [
                        match.player_tables[0].dice.colors.copy(),
                        match.player_tables[1].dice.colors.copy(),
                    ]
                    assert len(current_colors[0]) == 9
                    assert len(current_colors[1]) == 11
                    assert current_colors[0][0] == "OMNI"
                    assert current_colors[1][0] == "OMNI"
                    assert current_colors[1][1] == "OMNI"
                    assert current_colors[1][2] == "OMNI"
                    assert current_colors[0][1:] == last_colors[0]
                    assert current_colors[1][3:] == last_colors[1]
                    for support in match.player_tables[0].supports:
                        assert support.name != "Timmie"
                    assert len(match.player_tables[1].supports) == 0
                    assert len(match.player_tables[0].hands) == 10
                    assert len(match.player_tables[1].hands) == 10
                elif test_id == 22:
                    last_colors = [
                        match.player_tables[0].dice.colors.copy(),
                        match.player_tables[1].dice.colors.copy(),
                    ]
                else:
                    break
            make_respond(agent_1, match)
        else:
            raise AssertionError("No need respond.")
        if len(agent_1.commands) == 0:
            break

    assert match.state != MatchState.ERROR


def test_liben():
    cmd_records = [
        [
            "sw_card",
            "choose 0",
            "reroll 1 7",
            "card 0 0",
            "card 0 0",
            "sw_char 1 0",
            "sw_char 0 0",
            "sw_char 1 3",
            "sw_char 0 3",
            "end",
            "TEST 1 p0 usage 2 1",
            "TEST 1 p1 usage 3 3 2 0",
            "reroll",
            "sw_char 1 3",
            "sw_char 0 3",
            "sw_char 1 3",
            "sw_char 0 3",
            "sw_char 1 3",
            "end",
            "TEST 1 p0 usage 3 3",
            "reroll",
        ],
        [
            "sw_card",
            "choose 0",
            "reroll",
            "card 0 0",
            "card 0 0",
            "card 0 0",
            "card 0 0",
            "end",
            "reroll",
            "TEST 2 support all 2",
            "TEST 3 hand 7 dice 12 omni 4",
            "end",
            "reroll",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.0
        character:PyroMobMage
        character:ElectroMobMage
        character:Noelle
        Liben*30
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            # "TEST 2 support all 2",
            # "TEST 3 hand 7 dice 12 omni 4",
            elif test_id == 1:
                cmd = cmd.strip().split()
                pid = int(cmd[2][1])
                usg = [int(x) for x in cmd[4:]]
                for support, u in zip(match.player_tables[pid].supports, usg):
                    assert support.usage == u
            elif test_id == 2:
                for table in match.player_tables:
                    assert len(table.supports) == 2
            elif test_id == 3:
                table = match.player_tables[1]
                assert len(table.hands) == 7
                assert len(table.dice.colors) == 12
                for color in table.dice.colors[:4]:
                    assert color == "OMNI"
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_setaria():
    cmd_records = [
        [
            "sw_card",
            "choose 0",
            "reroll",
            "card 2 0 0",
            "card 3 0 0",
            "tune 0 0",
            "tune 0 0",
            "card 0 0 0",
            "TEST 1 p0 hand 2",
            "card 0 0 0",
            "tune 0 1",
            "TEST 1 p0 hand 1",
            "TEST 2 p0 support usage 2 3 3",
            "tune 0 2",
            "tune 0 3",
            "card 0 0 0",
            "TEST 2 p0 support usage 2 3 3",
            "card 0 0 0",
            "card 0 0 0",
            "card 0 0 0",
            "end",
            "TEST 2 p0 support usage 2 3 3",
            "reroll",
        ],
        ["sw_card", "choose 0", "reroll", "sw_char 1 0", "end"],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Fischl
        character:Mona
        character:Nahida
        Setaria*15
        Strategize*15
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                cmd = cmd.split()
                pid = int(cmd[2][1])
                assert len(match.player_tables[pid].hands) == int(cmd[4])
            elif test_id == 2:
                cmd = cmd.split()
                pid = int(cmd[2][1])
                us = [int(x) for x in cmd[5:]]
                assert len(match.player_tables[pid].supports) == len(us)
                for support, u in zip(match.player_tables[pid].supports, us):
                    assert support.usage == u
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_liusu():
    cmd_records = [
        [
            "sw_card",
            "choose 0",
            "card 0 0 15",
            "card 1 0 14",
            "card 1 0 13",
            "sw_char 1 12",
            "TEST 1 p0 support 1 2 2",
            "sw_char 2 11",
            "sw_char 1 10",
            "TEST 1 p0 support 1 1 2",
            "end",
            "sw_char 0 15",
            "sw_char 1 14",
        ],
        [
            "sw_card",
            "choose 1",
            "TEST 1 p0 support 1 2 2",
            "sw_char 2 15",
            "end",
            "card 1 0 15",
            "sw_char 1 14",
            "TEST 1 p0 support 1 2",
            "TEST 1 p1 support 1",
            "sw_char 0 13",
            "TEST 1 p0 support 1 2",
            "end",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Electro Hypostasis
        character:Klee
        character:Keqing
        Liu Su*30
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                # a sample of HP check based on the command string.
                cmd = cmd.split()
                pidx = int(cmd[2][1])
                usage = [int(x) for x in cmd[4:]]
                support = match.player_tables[pidx].supports
                assert len(usage) == len(support)
                for u, s in zip(usage, support):
                    assert u == s.usage
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_tubby():
    cmd_records = [
        [
            "sw_card 4",
            "choose 0",
            "reroll",
            "card 1 0 7 6",
            "card 0 0",
            "card 1 0",
            "reroll 2 3",
            "TEST 1 card 1 cost 2",
            "card 0 0 4 3",
            "card 0 3",
            "TEST 2 p0 dice 4",
            "end",
            "reroll",
            "reroll",
            "card 0 2",
            "card 0 1",
            "end",
            "reroll",
            "card 1 3",
            "end",
        ],
        [
            "sw_card 2 1",
            "choose 0",
            "reroll",
            "end",
            "reroll",
            "card 2 0 6 5",
            "card 1 0 4 3",
            "card 3 0",
            "card 1 0 2 1",
            "end",
            "reroll",
            "card 3 3",
            "reroll",
            "card 3 1 6 5",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Electro Hypostasis
        character:Klee
        character:Keqing
        Knights of Favonius Library*10
        Tubby*10
        Tenshukaku*10
        Vanarana*10
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                # a sample of HP check based on the command string.
                for req in match.requests:
                    if req.name == "UseCardRequest" and req.card_idx == 1:
                        assert req.cost.total_dice_cost == 2
            elif test_id == 2:
                assert len(match.player_tables[0].dice.colors) == 4
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_chang_nine():
    cmd_records = [
        [
            "sw_card",
            "choose 2",
            "card 2 0",
            "skill 0 15 14 13",
            "TEST 1 p0 support 1",
            "TEST 1 p1 support 1",
            "sw_char 4 12",
            "skill 2 11 10 9 8 7",
            "end",
            "card 5 4",
            "card 2 2",
            "card 2 0",
            "end",
            "TEST 1 p0 support 1",
            "TEST 1 p1 support 2",
            "skill 2 15 14 13 12 11",
        ],
        [
            "sw_card 1 2",
            "choose 0",
            "card 2 0",
            "skill 1 15 14 13",
            "sw_char 3 12",
            "TEST 1 p0 support 2",
            "TEST 1 p1 support 2",
            "skill 0 11 10 9",
            "card 1 0",
            "sw_char 1 8",
            "skill 0 7 6 5",
            "TEST 2 p0 card 6",
            "TEST 2 p1 card 5",
            "card 1 1",
            "TEST 1 p1 support 1",
            "sw_char 5 4",
            "skill 0 3 2 1",
            "sw_char 1 0",
            "end",
            "skill 0 15 14 13",
            "TEST 1 p0 support 1",
            "TEST 1 p1 support 2",
            "sw_char 2 12",
            "skill 1 11 10 9",
            "sw_char 3 8",
            "skill 1 7 6 5",
            "end",
            "card 0 0",
            "card 4 0",
            "card 6 0",
            "card 5 0",
            "TEST 1 p1 support 0 0 0 0",
            "card 4 1",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Xingqiu
        character:AnemoMobMage
        character:Yae Miko
        character:Mona
        character:Ganyu
        character:Klee
        Chang the Ninth*15
        Sweet Madame*15
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                cmd = cmd.split()
                pidx = int(cmd[2][1])
                usages = [int(x) for x in cmd[4:]]
                supports = match.player_tables[pidx].supports
                assert len(supports) == len(usages)
                for u, s in zip(usages, supports):
                    assert u == s.usage
            elif test_id == 2:
                cmd = cmd.split()
                pidx = int(cmd[2][1])
                card_num = int(cmd[4])
                assert len(match.player_tables[pidx].hands) == card_num
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_paimon_kujirai():
    json_fname = "kujirai_log.json"
    json_path = os.path.join(os.path.dirname(__file__), json_fname)
    do_log_tests(json_path, omnipotent=False)


def test_master_zhang():
    cmd_records = [
        [
            "sw_card 2 4",
            "choose 0",
            "end",
            "end",
            "card 1 0 15",
            "card 1 0 14",
            "TEST 1 card 1 cost 1",
            "card 1 0 13",
            "TEST 2 p0 support 0 0",
            "card 0 0",
            "TEST 1 card 5 cost 1",
            "card 4 0 12",
            "TEST 1 card 4 cost 0",
            "card 0 0 11",
            "card 3 1",
            "TEST 2 p0 support 0 0 0 1",
            "sw_char 1 10",
            "end",
            "card 3 0",
            "TEST 2 p0 support 0 0 1 1",
            "end",
        ],
        [
            "sw_card 1 2",
            "choose 0",
            "end",
            "end",
            "TEST 1 card 2 cost 3",
            "card 2 0 15 14 13",
            "card 4 0 12",
            "card 4 1 11",
            "end",
            "card 5 1",
            "TEST 1 card 6 cost 3",
            "card 5 0 15",
            "end",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Xingqiu
        character:Ganyu
        character:Fischl
        Master Zhang*15
        Where Is the Unseen Razor?*15
        King's Squire*15
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 2:
                cmd = cmd.split()
                pidx = int(cmd[2][1])
                usages = [int(x) for x in cmd[4:]]
                supports = match.player_tables[pidx].supports
                assert len(supports) == len(usages)
                for u, s in zip(usages, supports):
                    assert u == s.usage
            elif test_id == 1:
                cmd = cmd.split()
                cid = int(cmd[3])
                cost = int(cmd[5])
                for req in match.requests:
                    if req.name == "UseCardRequest" and req.card_idx == cid:
                        assert req.cost.total_dice_cost == cost
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def get_ellin_match():
    cmd_records = [
        [
            "sw_card 1 3",
            "choose 0",
            "skill 0 15 14 13",
            "sw_char 2 11",
            "skill 0 10 9 8",
            "card 3 0 7 6",
            "card 3 0 5 4",
            "skill 0 3 2 1",
            "end",
            "sw_char 1 0",
            "skill 0 0 1 2",
            "sw_char 2 0",
            "end",
            "TEST 1 p1 dice 10",
            "TEST 1 p0 dice 16",
            "card 0 0 15",
            "end",
            "sw_char 1 15",
            "sw_char 2 14",
            "skill 2 13 12 11 10",
            "card 0 0 9 8",
            "skill 0 7 6 5",
            "TEST 2 p1 support 0 0 0 1",
            "card 2 0 4 3",
            "TEST 3 skill 0 cost 1",
            "skill 0 2",
            "end",
        ],
        [
            "sw_card 1 3",
            "choose 2",
            "skill 0 15 14 13",
            "sw_char 0 11",
            "skill 0 10 9 8",
            "card 2 0 7 6",
            "end",
            "TEST 2 p0 support 1 1",
            "TEST 2 p1 support 1",
            "end",
            "TEST 2 p0 support 1",
            "TEST 4 p0c0 charge 2",
            "TEST 4 p0c1 charge 3",
            "TEST 4 p0c2 charge 2",
            "TEST 4 p1c0 charge 2",
            "TEST 4 p1c1 charge 1",
            "TEST 4 p1c2 charge 1",
            "card 2 0 15",
            "card 2 0 14 13",
            "sw_char 2 12",
            "sw_char 1 11",
            "sw_char 0 10",
            "sw_char 1 9",
            "TEST 2 p0 support 1 1",
            "card 0 0 8 7",
            "card 0 0 6 5",
            "card 3 1 4 3",
            "sw_char 2 2",
            "end",
            "TEST 1 p0 dice 14",
            "TEST 1 p1 dice 16",
            "card 4 0 15 14",
            "skill 1 12 11 10",
            "TEST 3 skill 0 cost 3",
            "TEST 3 skill 1 cost 0",
            "skill 0 9 8 7",
            "TEST 3 skill 0 cost 0",
            "TEST 3 skill 1 cost 0",
            "skill 0",
            "TEST 2 p0 support 1 0 0 0",
            "TEST 2 p1 support 0 0 0 1",
            "TEST 3 skill 0 cost 2",
            "TEST 3 skill 1 cost 2",
            "TEST 3 skill 2 cost 4",
            "end",
            "TEST 3 skill 0 cost 3",
            "end",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.1
        character:Kaedehara Kazuha
        character:Klee
        character:Kaeya
        Ellin*10
        Iron Tongue Tian*10
        Katheryne*5
        Katheryne@3.3*5
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    return agent_0, agent_1, match


def test_katheryne_tian_ellin():
    agent_0, agent_1, match = get_ellin_match()
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0].split()
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                pidx = int(cmd[2][1])
                assert len(match.player_tables[pidx].dice.colors) == int(cmd[4])
            elif test_id == 2:
                pidx = int(cmd[2][1])
                support = match.player_tables[pidx].supports
                check_usage(support, cmd[4:])
            elif test_id == 3:
                sid = int(cmd[3])
                cost = int(cmd[5])
                found = False
                for req in match.requests:
                    if req.name == "UseSkillRequest" and req.skill_idx == sid:
                        found = True
                        assert req.cost.total_dice_cost == cost
                assert found
            elif test_id == 4:
                pidx, cidx = get_pidx_cidx(cmd)
                assert match.player_tables[pidx].characters[cidx].charge == int(cmd[4])
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_wagner_timaeus():
    cmd_records = [
        [
            "sw_card",
            "choose 0",
            "card 0 0 15 14",
            "TEST 1 card 1 cost 1",
            "sw_char 2 13",
            "card 3 0 12 11",
            "end",
            "card 4 0 15 14",
            "TEST 2 p0 support 3 3 2",
            "end",
            "card 2 0 15 14",
            "TEST 1 card 4 cost 3",
            "card 3 0",
            "TEST 1 card 3 cost 0",
            "card 3 1",
            "TEST 2 p0 support 4 4 3 0",
            "end",
            "end",
            "end",
            "card 5 0 15 14",
            "card 0 2 13 12",
            "card 3 3 11 10",
            "card 3 3 9 8",
            "card 3 3 7 6",
            "end",
        ],
        [
            "sw_card 2 1 4",
            "choose 0",
            "card 2 0 15 14",
            "TEST 3 all card cost 0",
            "card 0 1",
            "TEST 2 p1 support 1",
            "TEST 3 all card cost 1",
            "card 0 1 13",
            "end",
            "card 2 0",
            "card 1 1",
            "card 0 1",
            "TEST 2 p1 usage 1",
            "TEST 1 card 0 usage 3",
            "end",
            "TEST 1 card 2 cost 3",
            "end",
            "card 1 0 15 14",
            "end",
            "card 0 1",
            "TEST 2 p1 usage 4 0",
            "end",
            "end",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.1
        character:Kaedehara Kazuha
        character:Klee
        character:Kaeya
        Timaeus*10
        Wagner*10
        Gambler's Earrings*10
        Tenacity of the Millelith*5
        Rhythm of the Great Dream*5
        Sacrificial Sword*5
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0].split()
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                cidx = int(cmd[3])
                cost = int(cmd[5])
                for req in match.requests:
                    if req.name == "UseCardRequest" and req.card_idx == cidx:
                        assert req.cost.total_dice_cost == cost
            elif test_id == 2:
                pidx = int(cmd[2][1])
                support = match.player_tables[pidx].supports
                check_usage(support, cmd[4:])
            elif test_id == 3:
                cost = int(cmd[-1])
                for req in match.requests:
                    if req.name == "UseCardRequest":
                        assert req.cost.total_dice_cost == cost
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_chef_mao():
    cmd_records = [
        [
            "sw_card",
            "choose 2",
            "skill 1 15 14 13",
            "sw_char 0 12",
            "skill 1 11 10 9",
            "card 2 0 8",
            "card 3 0 7",
            "card 1 0",
            "TEST 1 p0 card 4",
            "TEST 3 p0 dice 9",
            "end",
            "sw_char 2 15",
            "sw_char 1 14",
            "sw_char 2 13",
            "card 0 2",
            "TEST 1 p0 card 5",
            "TEST 3 p0 dice 15",
            "card 0 0",
            "card 0 0",
            "TEST 3 p0 dice 15",
            "card 1 0 14",
            "card 1 0 13",
            "end",
            "card 1 3 15",
            "card 0 2",
            "TEST 1 p0 card 3",
            "end",
            "card 3 0 15",
            "card 3 1 14",
            "card 1 1",
            "TEST 1 p0 card 4",
            "end",
            "card 1 1",
            "TEST 1 p0 card 5",
            "card 4 1 15",
            "card 2 0",
            "TEST 1 p0 card 3",
            "end",
        ],
        [
            "sw_card 1 2 3",
            "choose 2",
            "skill 1 15 14 13",
            "card 2 0",
            "TEST 1 p1 card 4",
            "TEST 3 p1 dice 13",
            "skill 1 12 11 10",
            "card 1 0 9",
            "card 1 1",
            "TEST 3 p1 dice 10",
            "TEST 1 p1 card 2",
            "end",
            "sw_char 1 15",
            "sw_char 0 14",
            "skill 0 13 12 11",
            "card 0 0 10",
            "card 0 1",
            "TEST 3 p1 dice 12",
            "TEST 1 p1 card 3",
            "sw_char 1 11",
            "end",
            "end",
            "end",
            "TEST 2 p0 support 0 0 0 0",
            "end",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.1
        character:Kaedehara Kazuha
        character:Klee
        character:Kaeya
        Chef Mao*10
        Chef Mao@3.3*10
        Sweet Madame*10
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0].split()
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                pidx = int(cmd[2][1])
                assert len(match.player_tables[pidx].hands) == int(cmd[-1])
            elif test_id == 2:
                pidx = int(cmd[2][1])
                support = match.player_tables[pidx].supports
                check_usage(support, cmd[4:])
            elif test_id == 3:
                pidx = int(cmd[2][1])
                assert len(match.player_tables[pidx].dice.colors) == int(cmd[-1])
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_dunyazard():
    cmd_records = [
        [
            "sw_card 0 1",
            "choose 0",
            "card 0 0 15",
            "TEST 1 card 0 cost 0",
            "TEST 1 card 1 cost 2",
            "TEST 1 card 2 cost 2",
            "end",
            "card 4 0 15 14",
            "end",
            "card 6 0",
            "end",
            "card 6 0",
            "card 4 1 15 14",
            "card 1 3 13 12",
            "end",
            "card 6 3 15 14",
            "end",
            "card 0 3",
            "end",
            "card 0 0",
            "TEST 3 p0 hand 9",
            "TEST 2 p0 support 0 0 1 2",
            "card 1 2",
            "TEST 3 p0 hand 8",
            "card 3 0",
            "TEST 3 p0 hand 8",
            "card 3 0 15 14",
            "end",
            "card 4 3 15 14",
            "card 7 3 13 12",
            "card 3 3",
            "TEST 3 p0 hand 7",
            "card 0 0 11",
            "TEST 3 p0 hand 7",
            "end",
            "end",
            "end",
            "end",
        ],
        [
            "sw_card 1 2",
            "choose 0",
            "TEST 1 card 1 cost 1",
            "card 1 0 15",
            "end",
            "card 5 0",
            "end",
            "card 5 0",
            "end",
            "card 0 0 15 14",
            "card 5 3",
            "end",
            "card 2 0",
            "TEST 2 p1 usage 0 0 1 2",
            "TEST 1 card 0 cost 2",
            "TEST 1 card 1 cost 0",
            "TEST 3 p1 hand 7",
            "card 5 3 15 14",
            "end",
            "card 1 3",
            "end",
            "card 0 0",
            "card 2 3",
            "TEST 3 p1 hand 8",
            "end",
            "card 3 0",
            "card 3 1",
            "end",
            "card 3 0 15 14",
            "end",
            "card 0 3",
            "TEST 3 p1 hand 10",
            "end",
            "TEST 4 p1 deck 11",
            "card 0 3",
            "card 0 3 15 14 13",
            "TEST 4 p1 deck 11",
            "end",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.1
        character:Kaedehara Kazuha
        character:Klee
        character:Kaeya
        Dunyarzad*10
        Dunyarzad@3.7*10
        Paimon*10
        NRE@3.3*10
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0].split()
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                cidx = int(cmd[3])
                cost = int(cmd[5])
                for req in match.requests:
                    if req.name == "UseCardRequest" and req.card_idx == cidx:
                        assert req.cost.total_dice_cost == cost
            elif test_id == 2:
                pidx = int(cmd[2][1])
                support = match.player_tables[pidx].supports
                check_usage(support, cmd[4:])
            elif test_id == 3:
                pidx = int(cmd[2][1])
                assert len(match.player_tables[pidx].hands) == int(cmd[-1])
            elif test_id == 4:
                pidx = int(cmd[2][1])
                assert len(match.player_tables[pidx].table_deck) == int(cmd[-1])
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_xudong():
    cmd_records = [
        [
            "sw_card",
            "choose 0",
            "card 2 0 15 14",
            "card 1 1",
            "card 3 1 13 12",
            "card 1 0 11 10",
            "card 3 0",
            "end",
            "choose 1",
            "card 2 0",
            "card 12 0",
            "TEST 1 card 0 cost 1",
            "card 5 0 15 14",
            "card 0 0",
            "TEST 2 p0 usage 0 0 0",
            "end",
        ],
        [
            "sw_card",
            "choose 0",
            "sw_char 2 15",
            "skill 1 14 13 12",
            "skill 1 11 10 9",
            "skill 2 8 7 6 5",
            "skill 1 4 3 2",
            "sw_char 1 1",
            "sw_char 2 0",
            "end",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.1
        character:Rhodeia of Loch
        character:Mona
        character:Kaeya
        Xudong*10
        Adeptus' Temptation*10
        Sweet Madame*10
        Lotus Flower Crisp*10
        Teyvat Fried Egg@3.7*10
        # Hanachirusato*10
        # Guardian's Oath*10
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    match.config.initial_hand_size = 20
    match.config.max_hand_size = 30
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0].split()
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                cidx = int(cmd[3])
                cost = int(cmd[5])
                for req in match.requests:
                    if req.name == "UseCardRequest" and req.card_idx == cidx:
                        assert req.cost.total_dice_cost == cost
            elif test_id == 2:
                pidx = int(cmd[2][1])
                support = match.player_tables[pidx].supports
                check_usage(support, cmd[4:])
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_hanachirusato():
    cmd_records = [
        [
            "sw_card",
            "choose 0",
            "skill 2 15 14 13 12 11",
            "card 3 0",
            "skill 0 10 9 8",
            "skill 0 7 6 5",
            "end",
            "card 2 0",
            "card 2 0",
            "TEST 1 card 0 cost 3",
            "card 2 0 15 14 13 12",
            "TEST 2 p0 usage 3 3 3",
            "card 2 0",
            "TEST 2 p0 usage 3",
            "card 0 0 11",
            "end",
            "TEST 2 p1 usage 3",
            "end",
        ],
        [
            "sw_card",
            "choose 0",
            "card 5 0",
            "card 3 0 15 14 13 12",
            "TEST 1 card 0 cost 3",
            "card 13 0",
            "TEST 2 p1 support 2 0",
            "skill 2 11 10 9 8 7",
            "end",
            "TEST 2 p1 usage 3 1",
            "TEST 2 p0 usage 1",
            "TEST 1 card 0 cost 1",
            "TEST 1 card 1 cost 4",
            "TEST 1 card 4 cost 0",
            "card 4 0",
            "TEST 2 p1 usage 1",
            "skill 2 15 14 13 12 11",
            "end",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.1
        character:Rhodeia of Loch
        character:Mona
        character:Noelle
        Hanachirusato*10
        Guardian's Oath*10
        Tenacity of the Millelith*10
        Gambler's Earrings*10
        The Bell*10
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    match.config.initial_hand_size = 20
    match.config.max_hand_size = 30
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0].split()
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                cidx = int(cmd[3])
                cost = int(cmd[5])
                for req in match.requests:
                    if req.name == "UseCardRequest" and req.card_idx == cidx:
                        assert req.cost.total_dice_cost == cost
            elif test_id == 2:
                pidx = int(cmd[2][1])
                support = match.player_tables[pidx].supports
                check_usage(support, cmd[4:])
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_chefmao_dunyarzad():
    cmd_records = [
        [
            "sw_card 2",
            "choose 0",
            "card 2 0 15",
            "TEST 1 p0 hand 4",
            "card 1 0",
            "TEST 1 p0 hand 4",
            "card 3 0",
            "TEST 1 p0 hand 3",
            "skill 0 14 13 12",
            "end",
            "end",
            "card 3 0",
            "TEST 1 p0 hand 7",
            "card 5 0",
            "TEST 1 p0 hand 7",
            "card 3 3",
            "TEST 1 p0 hand 6",
            "card 0 0",
            "TEST 1 p0 hand 7",
            "sw_char 1 15",
            "card 0 0",
            "TEST 1 p0 hand 6",
            "end",
        ],
        [
            "sw_card 2 1",
            "choose 0",
            "card 1 0 15",
            "card 3 0 14",
            "TEST 1 p1 hand 3",
            "card 1 0",
            "card 0 0",
            "TEST 1 p1 hand 3",
            "skill 0 13 15 14",
            "end",
            "end",
            "card 6 0 15",
            "card 0 0",
            "TEST 1 p1 hand 7",
            "card 4 3",
            "TEST 1 p1 hand 6",
            "card 1 0",
            "TEST 1 p1 hand 6",
            "skill 0 14 15 13",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.1
        character:Kaedehara Kazuha
        character:Klee
        character:Kaeya
        Dunyarzad*10
        Chef Mao*10
        Timmie*10
        Sweet Madame*10
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0].strip().split(" ")
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                pidx = int(cmd[2][1])
                hand = int(cmd[4])
                assert hand == len(match.player_tables[pidx].hands)
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_yayoi_fix():
    cmd_records = [
        ["sw_card", "choose 0", "end", "end", "end", "end"],
        [
            "sw_card 2 1",
            "choose 2",
            "card 1 0 15",
            "card 1 1 14 13",
            "end",
            "card 2 2 15",
            "end",
            "card 0 0",
            "end",
            "card 2 1",
            "card 4 2 15 14 13",
            "end",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.3
        character:Yoimiya@3.8
        character:Keqing@3.3
        character:Dehya@4.1
        Yayoi Nanatsuki*10
        Tenacity of the Millelith*10
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_timaeus_fix():
    cmd_records = [
        [
            "sw_card",
            "choose 0",
            "card 4 0 1",
            "card 5 0 0 1",
            "TEST 1 card 3 cost 0",
            "TEST 1 card 1 cost 0",
            "card 3 0",
            "TEST 1 card 1 cost 1",
            "end",
        ],
        ["sw_card 2 1 4", "choose 0", "end"],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.1
        character:Kaedehara Kazuha
        character:Klee
        character:Kaeya
        Strategize*10
        Timaeus*10
        Gambler's Earrings*10
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0].split()
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                cidx = int(cmd[3])
                cost = int(cmd[5])
                for req in match.requests:
                    if req.name == "UseCardRequest" and req.card_idx == cidx:
                        assert req.cost.total_dice_cost == cost
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_timaeus_fix_2():
    cmd_records = [
        [
            "sw_card 2 1",
            "choose 2",
            "card 3 0",
            "card 2 0 15 14",
            "card 1 0",
            "end",
            "sw_char 1 15",
            "card 0 0 14 13",
            "end",
        ],
        [
            "sw_card 3 2",
            "choose 2",
            "end",
            "card 1 0 15 14",
            "card 1 0",
            "card 1 0",
            "sw_char 1 13",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.2
        character:Baizhu
        character:Nilou
        character:Noelle
        Wagner*10
        Golden House*10
        The Bell*10
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def remove_ellin_information(match_dict: Dict):
    # remove Ellin information
    h = match_dict
    eh = h["event_handlers"]
    if len(eh) < 3:
        return json.dumps(match_dict)
    ellin_handler = h["event_handlers"][2]
    assert ellin_handler["name"] == "Ellin"
    ellin_handler["recorded_skill_ids"] = {}
    return json.dumps(h)


def test_ellin_load_and_save():
    agent_0, agent_1, match = get_ellin_match()
    assert match.start()[0]
    match.step()
    histories = [remove_ellin_information(match.dict())]
    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        while agent.commands[0].startswith("TEST"):
            agent.commands = agent.commands[1:]
        # respond
        make_respond(agent, match)
        histories.append(remove_ellin_information(match.dict()))
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break
    agent_0, agent_1, match = get_ellin_match()
    assert match.start()[0]
    match.step()
    for idx, old_match_json in enumerate(histories):  # pragma: no branch
        match_copy = match.copy(deep=True)
        assert remove_ids(match_copy) == remove_ids(
            Match(**json.loads(match_copy.json()))
        )
        assert remove_ids(
            Match(**json.loads(remove_ellin_information(match_copy.dict())))
        ) == remove_ids(Match(**json.loads(old_match_json)))
        match = Match(**json.loads(match.json()))
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        while agent.commands[0].startswith("TEST"):
            agent.commands = agent.commands[1:]
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break
    assert match.state != MatchState.ERROR


if __name__ == "__main__":
    # import logging
    # logging.basicConfig(level = logging.INFO)
    # test_rana()
    # test_timmie()
    # test_liben()
    # test_setaria()
    # test_liusu()
    # test_tubby()
    # test_chang_nine()
    test_paimon_kujirai()
    # test_master_zhang()
    # test_katheryne_tian_ellin()
    # test_wagner_timaeus()
    # test_chef_mao()
    # test_dunyazard()
    # test_xudong()
    # test_hanachirusato()
    # test_chefmao_dunyarzad()
    # test_yayoi_fix()
    # test_timaeus_fix_2()
    test_ellin_load_and_save()



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...action import (
    Actions,
    ChangeObjectUsageAction,
    CreateObjectAction,
    MakeDamageAction,
    RemoveObjectAction,
)
from ...event import (
    CreateObjectEventArguments,
    MakeDamageEventArguments,
    ReceiveDamageEventArguments,
    RoundPrepareEventArguments,
    SkillEndEventArguments,
)
from ...struct import Cost, ObjectPosition

from ...consts import (
    DamageElementalType,
    DamageType,
    ElementType,
    ElementalReactionType,
    IconType,
    ObjectPositionType,
)

from ...modifiable_values import DamageIncreaseValue, DamageValue
from .base import RoundTeamStatus, ShieldTeamStatus, SwitchActionTeamStatus


class ShrineOfMaya_3_7(RoundTeamStatus):
    name: Literal["Shrine of Maya"] = "Shrine of Maya"
    version: Literal["3.7"] = "3.7"
    usage: int = 2
    max_usage: int = 2
    icon_type: Literal[IconType.OTHERS] = IconType.OTHERS

    def event_handler_CREATE_OBJECT(
        self, event: CreateObjectEventArguments, match: Any
    ) -> List[ChangeObjectUsageAction]:
        """
        When self is created, check whether to increase the usage of Seed of
        Skandha.
        """
        if event.action.object_name != "Shrine of Maya":
            # not creating Shrine of Maya, do nothing.
            return []
        active_character = match.player_tables[
            self.position.player_idx
        ].get_active_character()
        assert active_character is not None
        if active_character.name == "Nahida" and active_character.talent is not None:
            # talent equipped, check if have electro character.
            has_electro_character = False
            for character in match.player_tables[self.position.player_idx].characters:
                if character.element == ElementType.ELECTRO:
                    has_electro_character = True
                    break
            if has_electro_character:
                # has electro character, add one usage for enemy
                # Seed of Skandha.
                ret = []
                for character in match.player_tables[
                    1 - self.position.player_idx
                ].characters:
                    for status in character.status:
                        if status.name == "Seed of Skandha":
                            position = status.position.set_area(
                                ObjectPositionType.CHARACTER_STATUS
                            )
                            ret.append(
                                ChangeObjectUsageAction(
                                    object_position=position,
                                    change_usage=1,
                                )
                            )
                return ret
        return []

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any, mode: Literal["TEST", "REAL"]
    ) -> DamageIncreaseValue:
        """
        +1 on self elemental reaction damage.
        """
        if value.damage_type != DamageType.DAMAGE:
            # not damage, not modify
            return value
        if value.target_position.player_idx == self.position.player_idx:
            # attack self, not activate
            return value
        if value.position.player_idx != self.position.player_idx:
            # not self, not activate
            return value
        if value.element_reaction is not ElementalReactionType.NONE:
            value.damage += 1
        return value


class FloralSidewinder_3_3(RoundTeamStatus):
    """
    Damage made on skill end, but need to check whether dendro reaction made
    before.
    """

    name: Literal["Floral Sidewinder"] = "Floral Sidewinder"
    version: Literal["3.3"] = "3.3"
    usage: int = 1
    max_usage: int = 1
    dendro_reaction_made: bool = False
    icon_type: Literal[IconType.ATK_SELF] = IconType.ATK_SELF

    def event_handler_RECEIVE_DAMAGE(
        self, event: ReceiveDamageEventArguments, match: Any
    ) -> List[Actions]:
        """
        Check whether dendro reaction made.
        """
        damage = event.final_damage
        if not damage.is_corresponding_character_use_damage_skill(
            self.position, match, None
        ):
            # not self character made damage
            return []
        if ElementType.DENDRO not in damage.reacted_elements:
            # not dendro reaction
            return []
        # mark dendro reaction made
        self.dendro_reaction_made = True
        return []

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[MakeDamageAction | RemoveObjectAction]:
        """
        if no dendro reaction made, return. otherwise, reset dendro reaction
        made, and if our active character use skill, deal 1 dendro damage.
        """
        if not self.dendro_reaction_made:
            # no dendro reaction made
            return []
        # regardless of the following conditions, reset dendro reaction made
        self.dendro_reaction_made = False
        character = match.player_tables[self.position.player_idx].get_active_character()
        if not event.action.position.check_position_valid(  # pragma: no cover
            character.position,
            match,
            player_idx_same=True,
            character_idx_same=True,
            source_area=ObjectPositionType.SKILL,
        ):
            # not this character use skill
            return []  # pragma: no cover
        target = match.player_tables[
            1 - self.position.player_idx
        ].get_active_character()
        self.usage -= 1
        return [
            MakeDamageAction(
                damage_value_list=[
                    DamageValue(
                        position=self.position,
                        damage_type=DamageType.DAMAGE,
                        target_position=target.position,
                        damage=1,
                        damage_elemental_type=DamageElementalType.DENDRO,
                        cost=Cost(),
                    )
                ]
            )
        ] + self.check_should_remove()


class AdeptalLegacy_4_1(SwitchActionTeamStatus):
    name: Literal["Adeptal Legacy"] = "Adeptal Legacy"
    version: Literal["4.1"] = "4.1"
    usage: int = 3
    max_usage: int = 3
    icon_type: Literal[IconType.OTHERS] = IconType.OTHERS

    def _action(self, match: Any) -> List[Actions]:
        """
        attack enemy active character and heal our active character.
        """
        enemy_active_character = match.player_tables[
            1 - self.position.player_idx
        ].get_active_character()
        our_active_character = match.player_tables[
            self.position.player_idx
        ].get_active_character()
        return [
            MakeDamageAction(
                damage_value_list=[
                    DamageValue(
                        position=self.position,
                        damage_type=DamageType.DAMAGE,
                        target_position=enemy_active_character.position,
                        damage=1,
                        damage_elemental_type=DamageElementalType.DENDRO,
                        cost=Cost(),
                    )
                ]
            ),
            MakeDamageAction(
                damage_value_list=[
                    DamageValue(
                        position=self.position,
                        damage_type=DamageType.HEAL,
                        target_position=our_active_character.position,
                        damage=-1,
                        damage_elemental_type=DamageElementalType.HEAL,
                        cost=Cost(),
                    )
                ]
            ),
        ]


class PulsingClarity_4_2(RoundTeamStatus):
    name: Literal["Pulsing Clarity"] = "Pulsing Clarity"
    version: Literal["4.2"] = "4.2"
    usage: int = 2
    max_usage: int = 2
    icon_type: Literal[IconType.OTHERS] = IconType.OTHERS

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Any
    ) -> List[Actions]:
        ret = super().event_handler_ROUND_PREPARE(event, match)
        return [
            CreateObjectAction(
                object_name="Seamless Shield",
                object_position=ObjectPosition(
                    player_idx=self.position.player_idx,
                    area=ObjectPositionType.TEAM_STATUS,
                    id=0,
                ),
                object_arguments={},
            )
        ] + ret


class SeamlessShield_4_2(ShieldTeamStatus):
    name: Literal["Seamless Shield"] = "Seamless Shield"
    version: Literal["4.2"] = "4.2"
    usage: int = 1
    max_usage: int = 1

    def effect_action(self, match: Any) -> MakeDamageAction:
        """
        make damage and heal
        """
        our_active_character = match.player_tables[
            self.position.player_idx
        ].get_active_character()
        opposite_active_character = match.player_tables[
            1 - self.position.player_idx
        ].get_active_character()
        ret = MakeDamageAction(
            damage_value_list=[
                DamageValue(
                    position=self.position,
                    damage_type=DamageType.DAMAGE,
                    target_position=opposite_active_character.position,
                    damage=1,
                    damage_elemental_type=DamageElementalType.DENDRO,
                    cost=Cost(),
                ),
            ],
        )
        if our_active_character.hp > 0:
            # when active character has no hp, no need to heal
            ret.damage_value_list.append(
                DamageValue(
                    position=self.position,
                    damage_type=DamageType.HEAL,
                    target_position=our_active_character.position,
                    damage=-1,
                    damage_elemental_type=DamageElementalType.HEAL,
                    cost=Cost(),
                ),
            )
        return ret

    def event_handler_CREATE_OBJECT(
        self, event: CreateObjectEventArguments, match: Any
    ) -> List[MakeDamageAction]:
        if not self.position.check_position_valid(
            event.action.object_position,
            match,
            player_idx_same=True,
            area_same=True,
        ):
            # not create on our area
            return []
        if event.action.object_name != self.name or event.create_result != "RENEW":
            # not create ourself, or not renew create
            return []
        # renew, make damage and heal
        return [self.effect_action(match)]

    def event_handler_MAKE_DAMAGE(
        self, event: MakeDamageEventArguments, match: Any
    ) -> List[MakeDamageAction | RemoveObjectAction]:
        """
        When damage made, check whether the team status should be removed.
        """
        ret = self.check_should_remove()
        if len(ret) > 0:
            return [
                self.effect_action(match),
                RemoveObjectAction(object_position=self.position),
            ]
        return list(ret)


register_class(
    ShrineOfMaya_3_7
    | FloralSidewinder_3_3
    | AdeptalLegacy_4_1
    | PulsingClarity_4_2
    | SeamlessShield_4_2
)



from typing import Any, List, Literal

from .....utils.class_registry import register_class

from ....consts import ELEMENT_TO_DIE_COLOR, ElementalReactionType
from ....action import CreateDiceAction
from ....event import ReceiveDamageEventArguments, SkillEndEventArguments
from ....struct import Cost
from .base import RoundEffectArtifactBase


class InstructorsCap_3_3(RoundEffectArtifactBase):
    # TODO electro hypostasis triggers reaction for second punch,
    # may generate dice for the next attack. Same as Chang and parametric?
    name: Literal["Instructor's Cap"]
    version: Literal["3.3"] = "3.3"
    cost: Cost = Cost(any_dice_number=2)
    max_usage_per_round: int = 3
    element_reaction_triggered: bool = False

    def event_handler_RECEIVE_DAMAGE(
        self, event: ReceiveDamageEventArguments, match: Any
    ) -> List[CreateDiceAction]:
        """
        Record if elemental reaction triggered
        """
        if event.elemental_reaction != ElementalReactionType.NONE:
            # has elemental reaction
            self.element_reaction_triggered = True
        return []

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[CreateDiceAction]:
        """
        if elemental reaction reiggered, and self use skill, and have usage,
        create one die with same element type as self.
        """
        if not self.element_reaction_triggered:
            # no elemental reaction triggered
            return []
        self.element_reaction_triggered = False
        if self.usage <= 0:
            # no usage
            return []
        if self.position.not_satisfy(
            "both pidx=same cidx=same and source area=character and target area=skill",
            event.action.position,
        ):
            # not self player use skill or not equipped
            return []
        # create die
        self.usage -= 1
        character = match.player_tables[self.position.player_idx].characters[
            self.position.character_idx
        ]
        return [
            CreateDiceAction(
                player_idx=self.position.player_idx,
                number=1,
                color=ELEMENT_TO_DIE_COLOR[character.element],
            )
        ]


register_class(InstructorsCap_3_3)



from lpsim.server.deck import Deck
from lpsim.server.interaction import UseCardResponse
from lpsim.server.match import Match, MatchState

from tests.utils_for_test import (
    check_hp,
    get_random_state,
    get_test_id_from_command,
    make_respond,
    set_16_omni,
)
from lpsim.agents.interaction_agent import InteractionAgent
from lpsim.agents.nothing_agent import NothingAgent


def test_covenant_of_rock():
    agent_0 = NothingAgent(player_idx=0)
    agent_1 = InteractionAgent(
        player_idx=1,
        verbose_level=0,
        commands=[
            "sw_card",
            "choose 0",
            "reroll",
            "TEST 1 no card can use",
            "sw_char 1 0",
            "tune 0 6",
            "skill 0 0 1 2",
            "tune 0 1",
            "skill 0 0 1 2",
            "sw_char 2 0",
            "TEST 2 can use 3 card",
            "card 0 0",
            "TEST 3 dice 2 not same no omni",
            "sw_char 1 0",
            "sw_char 0 0",
            "TEST 1 no card can use",
            "end",
            "reroll",
            "TEST 1 no card can use",
            "TEST 4 table arcane still false",
            "sw_char 1 1",
        ],
        only_use_command=True,
    )
    match = Match(random_state=get_random_state())
    deck = Deck.from_str(
        """
        character:Fischl
        character:Mona
        character:Nahida
        # Wine-Stained Tricorne*2
        # Timmie*2
        # Rana*2
        # Strategize*2
        # The Bestest Travel Companion!*2
        # Covenant of Rock
        Covenant of Rock*30
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = 30
    assert match.start()[0]
    match.step()

    while True:
        if match.need_respond(0):
            make_respond(agent_0, match)
        elif match.need_respond(1):
            while True:
                test_id = get_test_id_from_command(agent_1)
                if test_id == 1:
                    for request in match.requests:
                        assert request.name != "UseCardRequest"
                elif test_id == 2:
                    card_req = 0
                    for request in match.requests:
                        if request.name == "UseCardRequest":
                            card_req += 1
                    assert card_req == 3
                elif test_id == 3:
                    colors = match.player_tables[1].dice.colors.copy()
                    colorset = set(colors)
                    assert len(colors) == len(colorset)
                    assert len(colors) == 2
                    for color in colors:
                        assert color != "OMNI"
                elif test_id == 4:
                    assert not match.player_tables[1].arcane_legend
                else:
                    break
            make_respond(agent_1, match)
        else:
            raise AssertionError("No need respond.")
        if len(agent_1.commands) == 0:
            break

    assert match.state != MatchState.ERROR


def test_rock_dice_different_not_omni():
    """
    stop in using card, and set random random_state, then use card again
    to check always different and not omni
    """
    agent_0 = NothingAgent(player_idx=0)
    agent_1 = InteractionAgent(
        player_idx=1,
        verbose_level=0,
        commands=[
            "sw_card",
            "choose 0",
            "reroll",
            "sw_char 1 0",
            "tune 0 6",
            "skill 0 0 1 2",
            "tune 0 1",
            "skill 0 0 1 2",
            "sw_char 2 0",
        ],
        only_use_command=True,
    )
    match = Match(random_state=get_random_state())
    deck = Deck.from_str(
        """
        character:Fischl
        character:Mona
        character:Nahida
        # Wine-Stained Tricorne*2
        # Timmie*2
        # Rana*2
        # Strategize*2
        # The Bestest Travel Companion!*2
        # Covenant of Rock
        Covenant of Rock*30
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = 30
    assert match.start()[0]
    match.step()

    while True:
        if match.need_respond(0):
            make_respond(agent_0, match)
        elif match.need_respond(1):
            make_respond(agent_1, match)
        else:
            raise AssertionError("No need respond.")
        if len(agent_1.commands) == 0:
            break
    match.step()

    assert match.state != MatchState.ERROR
    assert match.need_respond(1)
    card_req = None
    for req in match.requests:
        if req.name == "UseCardRequest":
            card_req = req
    assert card_req is not None
    card_resp = UseCardResponse(request=card_req, dice_idxs=[], target=None)
    match_bk = match.copy(deep=True)
    for _ in range(50):
        match = match_bk.copy(deep=True)
        match_new = Match()
        match.random_state = match_new.random_state
        match._init_random_state()
        match.respond(card_resp)
        match.step()
        colors = match.player_tables[1].dice.colors.copy()
        colorset = set(colors)
        assert len(colors) == len(colorset)
        for color in colors:
            assert color != "OMNI"
        assert len(colors) == 2


def test_arcane_card_always_in_hand():
    """
    when 3 arcane card, they should always be the first three
    """
    deck = Deck.from_str(
        """
        character:Fischl
        character:Mona
        character:Nahida
        # Wine-Stained Tricorne*2
        # Timmie*2
        # Rana*2
        # Strategize*2
        # The Bestest Travel Companion!*2
        # Covenant of Rock
        Covenant of Rock*3
        Strategize*27
        """
    )
    for i in range(50):
        match = Match()
        match.set_deck([deck, deck])
        match.config.max_same_card_number = 30
        assert match.start()[0]
        match.step()
        for table in match.player_tables:
            first_three = table.hands[:3]
            for card in first_three:
                assert card.name == "Covenant of Rock"


def test_ancient_courtyard():
    cmd_records = [
        [
            "sw_card 0 1 2 3",
            "choose 1",
            "reroll",
            "TEST 1 card 0 cannot use",
            "card 1 0 1 0",
            "TEST 2 card 0 can use",
            "card 1 0",
            "TEST 1 card 0 cannot use",
            "card 1 0",
            "sw_char 2 5",
            "end",
            "reroll",
            "end",
            "reroll",
            "end",
            "reroll",
            "card 0 0",
            "sw_char 1 7",
            "card 5 1",
            "TEST 3 p0 no team status",
            "end",
        ],
        [
            "sw_card 0 1 2 3",
            "choose 0",
            "reroll",
            "TEST 1 card 0 cannot use",
            "card 1 2 3",
            "TEST 2 card 0 can use",
            "card 1 1 6 5",
            "card 1 0",
            "card 0 0",
            "card 0 1",
            "card 0 0",
            "skill 0 0 4 2",
            "end",
            "reroll",
            "end",
            "reroll",
            "end",
            "reroll",
            "TEST 4 card 5 cost 2",
            "TEST 4 card 1 cost 1",
            "card 1 1 4",
            "card 4 2 6 5",
            "end",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Mona
        character:Nahida
        character:Klee
        Ancient Courtyard*10
        Where Is the Unseen Razor?*10
        Gambler's Earrings*10
        Magic Guide*10
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                # a sample of HP check based on the command string.
                cmd = cmd.split()
                cardnum = int(cmd[3])
                for req in match.requests:
                    if req.name == "UseCardRequest":
                        assert req.card_idx != cardnum
            elif test_id == 2:
                cmd = cmd.split()
                cardnum = int(cmd[3])
                found = False
                for req in match.requests:
                    if req.name == "UseCardRequest" and req.card_idx == cardnum:
                        found = True
                assert found
            elif test_id == 3:
                assert len(match.player_tables[0].team_status) == 0
            elif test_id == 4:
                cmd = cmd.split()
                cardnum = int(cmd[3])
                costnum = int(cmd[5])
                for req in match.requests:
                    if req.name == "UseCardRequest" and req.card_idx == cardnum:
                        assert req.cost.total_dice_cost == costnum
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_joyous():
    cmd_records = [
        [
            "sw_card",
            "choose 0",
            "skill 1 15 14 13",
            "TEST 1 card 0 can use",
            "sw_char 3 12",
            "TEST 2 card 0 cannot use",
            "sw_char 4 11",
            "TEST 2 card 0 cannot use",
            "sw_char 1 10",
            "sw_char 2 9",
            "skill 0 8 7 6",
            "sw_char 3 5",
            "end",
            "sw_char 0 15",
            "sw_char 2 14",
            "card 0 0 0",
            "TEST 3 p1 summon usage 1",
            "TEST 4 p0c0 pyro",
            "TEST 4 p0c4 pyro",
            "TEST 4 p0c1 none",
            "TEST 4 p0c2 none",
            "TEST 4 p0c3 none",
            "end",
        ],
        [
            "sw_card",
            "choose 2",
            "skill 1 15 14 13",
            "sw_char 1 12",
            "card 0 0 0",
            "TEST 5 p0 team dendro core",
            "TEST 6 p1 all dendro except 2",
            "skill 0 1 10 9",
            "sw_char 0 1",
            "skill 1 1 6 5",
            "end",
            "skill 0 15 14 13",
            "sw_char 2 12",
            "skill 0 11 10",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Mona
        character:Nahida
        character:Klee
        character:Noelle
        character:Venti
        Joyous Celebration
        Where Is the Unseen Razor?*10
        Gambler's Earrings*10
        Magic Guide*10
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                # a sample of HP check based on the command string.
                cmd = cmd.split()
                found = False
                for req in match.requests:
                    if req.name == "UseCardRequest" and req.card_idx == 0:
                        found = True
                assert found
            elif test_id == 2:
                cmd = cmd.split()
                for req in match.requests:
                    if req.name == "UseCardRequest":
                        assert req.card_idx != 0
            elif test_id == 3:
                assert len(match.player_tables[1].summons) == 1
                assert match.player_tables[1].summons[0].usage == 1
            elif test_id == 4:
                cmd = cmd.split()
                pidx = int(cmd[2][1])
                cidx = int(cmd[2][3])
                character = match.player_tables[pidx].characters[cidx]
                if cmd[3] == "none":
                    assert character.element_application == []
                else:
                    assert character.element_application == [cmd[3].upper()]
            elif test_id == 5:
                team_status = match.player_tables[0].team_status
                assert len(team_status) == 1
                assert team_status[0].name == "Dendro Core"
            elif test_id == 6:
                characters = match.player_tables[1].characters
                for cid, c in enumerate(characters):
                    if cid == 2:
                        assert c.element_application == []
                    else:
                        assert c.element_application == ["DENDRO"]
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_joyous_2():
    cmd_records = [
        [
            "sw_card",
            "choose 0",
            "skill 1 15 14 13",
            "sw_char 2 12",
            "skill 1 11 10 9",
            "skill 1 8 7 6",
            "skill 0 5 4",
        ],
        [
            "sw_card",
            "choose 2",
            "skill 1 15 14 13",
            "skill 1 12 11 10",
            "skill 0 9 8",
            "choose 0",
            "card 0 0 0",
            "TEST 6 p1 all hydro except 2",
            "end",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Mona
        character:Nahida
        character:Klee
        character:Noelle
        character:Venti
        Joyous Celebration
        Where Is the Unseen Razor?*10
        Gambler's Earrings*10
        Magic Guide*10
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 6:
                characters = match.player_tables[1].characters
                for cid, c in enumerate(characters):
                    if cid == 2:
                        assert c.element_application == []
                    else:
                        assert c.element_application == ["HYDRO"]
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_fresh_wind_of_freedom():
    cmd_records = [
        [
            "sw_card",
            "choose 0",
            "skill 0 15 14 13",
            "skill 0 12 11 10",
            "end",
            "skill 0 15 14 13",
            "choose 2",
            "TEST 1 0 10 8 10 10 4",
            "end",
            "card 3 0",
            "skill 0 15 14 13",
            "choose 1",
            "TEST 1 0 10 0 8 10 4",
            "skill 1 12 11 10",
            "TEST 1 0 8 0 5 10 4",
            "end",
        ],
        [
            "sw_card",
            "choose 2",
            "skill 1 15 14 13",
            "skill 1 12 11 10",
            "skill 2 9 8 7 6",
            "skill 0 5 4 3",
            "end",
            "card 2 0",
            "sw_char 1 15",
            "sw_char 0 14",
            "skill 0 13 12 11",
            "skill 0 10 9 8",
            "skill 0 7 6 5",
            "end",
            "skill 0 15 14 13",
            "sw_char 1 12",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.1
        character:Kaedehara Kazuha
        character:Klee
        character:Kaeya
        Fresh Wind of Freedom*10
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                # a sample of HP check based on the command string.
                hps = cmd.strip().split(" ")[2:]
                hps = [int(x) for x in hps]
                hps = [hps[:3], hps[3:]]
                check_hp(match, hps)
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_arcaneguoba_hairan_fenglong_lyresong():
    cmd_records = [
        [
            "sw_card 1 2",
            "choose 0",
            "skill 1 15 14 13",
            "TEST 1 7 8 8 10 10 10",
            "card 0 0",
            "sw_char 1 12",
            "card 2 1 11 10 9",
            "card 0 0 8 7",
            "card 1 0 6 5",
            "TEST 2 skill 2 cost 3",
            "TEST 3 card 1 cost 1",
            "TEST 4 p0 hand 4",
            "card 1 0 4",
            "sw_char 1 3",
            "skill 0 2 1 0",
            "TEST 1 7 10 8 10 5 10",
            "end",
            "TEST 1 7 7 8 10 10 10",
            "sw_char 0 15",
            "card 0 0",
            "card 4 0 14",
            "skill 0 13 12 11",
            "TEST 1 7 4 5 10 10 6",
            "card 0 0",
            "card 3 1 10",
            "sw_char 1 9",
            "end",
            "sw_char 0 15",
            "TEST 5 p1 support",
            "skill 2 14 13",
            "sw_char 2 12",
            "end",
            "card 0 0 15 14 13",
            "TEST 5 p0 support",
            "TEST 1 4 7 5 8 6 7",
            "end",
        ],
        [
            "sw_card 2 3 4",
            "choose 2",
            "TEST 1 10 10 10 10 10 7",
            "card 4 2 15 14 13",
            "card 1 0 12 11",
            "TEST 2 skill 2 cost 4",
            "TEST 3 card 3 cost 4",
            "card 3 0 10 9 8 7",
            "sw_char 1 6",
            "TEST 2 skill 2 cost 5",
            "end",
            "card 4 1 15 14 13",
            "card 1 0",
            "card 3 1 12",
            "skill 2 11 10 9 8",
            "sw_char 2 7",
            "skill 2 6 5 4 3 2",
            "sw_char 1 1",
            "end",
            "skill 2 15 14 13 12",
            "card 2 1",
            "card 4 2 11",
            "card 3 0",
            "card 2 1 10",
            "sw_char 0 9",
            "end",
            "card 1 1",
            "card 4 2 15",
            "card 3 0 14 13 12",
            "card 2 0 11 10",
            "end",
            "card 0 0",
            "TEST 4 p1 hand 7",
            "end",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.2
        character:Keqing
        character:Nahida
        character:Ganyu
        Lyresong
        Stormterror's Lair*10
        Ocean-Hued Clam*10
        Lyresong*9
        In Every House a Stove
        Undivided Heart
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0].strip().split(" ")
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                # a sample of HP check based on the command string.
                hps = cmd[2:]
                hps = [int(x) for x in hps]
                hps = [hps[:3], hps[3:]]
                check_hp(match, hps)
            elif test_id == 2:
                sidx = int(cmd[3])
                cost = int(cmd[5])
                for req in match.requests:
                    if req.name == "UseSkillRequest" and req.skill_idx == sidx:
                        assert req.cost.total_dice_cost == cost
            elif test_id == 3:
                cidx = int(cmd[3])
                cost = int(cmd[5])
                for req in match.requests:
                    if req.name == "UseCardRequest" and req.card_idx == cidx:
                        assert req.cost.total_dice_cost == cost
            elif test_id == 4:
                pidx = int(cmd[2][1])
                hnum = int(cmd[4])
                assert len(match.player_tables[pidx].hands) == hnum
            elif test_id == 5:
                pidx = int(cmd[2][1])
                assert len(match.player_tables[pidx].supports) == 0
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


if __name__ == "__main__":
    # test_covenant_of_rock()
    test_rock_dice_different_not_omni()
    # test_arcane_card_always_in_hand()
    test_ancient_courtyard()
    test_joyous()
    test_fresh_wind_of_freedom()
    test_arcaneguoba_hairan_fenglong_lyresong()



import logging

from pydantic import validator

from ..utils.deck_code import deck_code_to_deck_str, deck_str_to_deck_code

from .object_base import CardBase

from .character.character_base import CharacterBase

from ..utils import BaseModel, get_instance
from typing import Any, Literal, List, Tuple


class Deck(BaseModel):
    name: Literal["Deck"] = "Deck"
    characters: List[CharacterBase] = []
    cards: List[CardBase] = []
    default_version: str | None = None

    @validator("characters", each_item=True, pre=True)
    def parse_characters(cls, v):
        return get_instance(CharacterBase, v)

    @validator("cards", each_item=True, pre=True)
    def parse_cards(cls, v):
        return get_instance(CardBase, v)

    def check_legal(
        self,
        card_number: int | None,
        max_same_card_number: int | None,
        character_number: int | None,
        check_restriction: bool,
    ) -> Tuple[bool, Any]:
        """
        check whether the deck is legal.
        1. the number of characters
        2. the number of cards
        3. the number of cards with the same name
        4. cards with special carrying rules

        Returns:
            bool: whether the deck is legal
            Any: error message if not legal
        """
        if character_number is not None:
            if len(self.characters) != character_number:
                error_msg = f"character number should be {character_number}"
                logging.error(error_msg)
                return False, error_msg
        if card_number is not None:
            if len(self.cards) != card_number:
                error_msg = f"card number should be {card_number}"
                logging.error(error_msg)
                return False, error_msg
        if max_same_card_number is not None:
            card_names = [c.name for c in self.cards]
            for card in card_names:
                if card_names.count(card) > max_same_card_number:
                    error_msg = (
                        f"card {card} should not be more than "
                        f"{max_same_card_number} in deck"
                    )
                    logging.error(error_msg)
                    return False, error_msg
        if check_restriction:
            for card in self.cards:
                restriction = card.get_deck_restriction()
                if restriction.type == "NONE":
                    continue
                elif restriction.type == "FACTION":
                    counter = 0
                    for character in self.characters:
                        if restriction.name in character.faction:
                            counter += 1
                    if counter < restriction.number:
                        error_msg = (
                            f"to use card {card.name}, "
                            f"character number of faction {restriction.name} "
                            f"should be at least {restriction.number}"
                        )
                        logging.error(error_msg)
                        return False, error_msg
                elif restriction.type == "CHARACTER":
                    counter = 0
                    for character in self.characters:
                        if restriction.name == character.name:
                            counter += 1
                    if counter < restriction.number:
                        error_msg = (
                            f"to use card {card.name}, "
                            f"character number of name {restriction.name} "
                            f"should be at least {restriction.number}"
                        )
                        logging.error(error_msg)
                        return False, error_msg
                elif restriction.type == "ELEMENT":
                    counter = 0
                    for character in self.characters:
                        if restriction.name == character.element.value:
                            counter += 1
                    if counter < restriction.number:
                        error_msg = (
                            f"to use card {card.name}, "
                            f"character number of element {restriction.name} "
                            f"should be at least {restriction.number}"
                        )
                        logging.error(error_msg)
                        return False, error_msg
                else:
                    raise NotImplementedError(
                        f"restriction type {restriction.type} not implemented"
                    )
        return True, None

    @staticmethod
    def from_str(deck_str: str) -> "Deck":
        """
        Convert deck string to deck object. One line contains one command,
        card or character.

        To specify default versions for all cards that don't have version
        specified, use 'default_version:version'. If not set, cards without
        version information will not have version in its argument. This line
        must appear before any other lines.

        If declare a character, use 'character:'. Otherwise
        write card name directly.

        To specify different version of cards, add `@version` after the
        line.

        To declare multiple same cards or characters, add `*number` after
        the line. When both marks are used, specify version first.
        You can write comment line start with '#'.

        Examples:

        default_version:4.0
        character:Fischl
        character:Mona
        character:Nahida
        # you can write comment line with '#'
        Timmie*15
        Rana*15

        """
        deck_str = deck_str.strip()
        deck = Deck()
        for line in deck_str.split("\n"):
            line = line.strip()
            if line == "":
                continue
            if line.startswith("default_version:"):
                deck.default_version = line[16:]
                continue
            if line[0] == "#":
                # comment line
                continue
            # deal with numbers
            if "*" in line:
                line, number = line.split("*")
                number = int(number)
            else:
                number = 1
            # deal with version
            if "@" in line:
                line, version = line.split("@")
                version = version.strip()
            else:
                version = deck.default_version
            if line.startswith("character:"):
                args = {"name": line[10:]}
                if version is not None:
                    args["version"] = version
                for _ in range(number):
                    deck.characters.append(get_instance(CharacterBase, args))
            else:
                for _ in range(number):
                    args = {"name": line}
                    if version is not None:
                        args["version"] = version
                    deck.cards.append(get_instance(CardBase, args))
        return deck

    def to_str(self) -> str:
        """
        Convert deck object to deck string.
        """
        deck_str = ""
        if self.default_version is not None:
            deck_str += f"default_version:{self.default_version}\n"
        for character in self.characters:
            deck_str += f"character:{character.name}"
            deck_str += f"@{character.version}"
            deck_str += "\n"
        for card in self.cards:
            deck_str += card.name
            deck_str += f"@{card.version}"
            deck_str += "\n"
        return deck_str

    @staticmethod
    def from_deck_code(deck_code: str, version: str | None = None) -> "Deck":
        """
        Convert deck code to deck object. If version is specified,
        default_version will be set in deck_str.
        """
        return Deck.from_str(deck_code_to_deck_str(deck_code, version))

    def to_deck_code(self) -> str:
        """
        Convert deck object to deck code.
        """
        return deck_str_to_deck_code(self.to_str())



import pytest

from lpsim.server.character.character_base import SkillBase
from lpsim.server.consts import SkillType
from lpsim.server.deck import Deck
from lpsim.server.match import Match, MatchState
from lpsim.agents.interaction_agent import InteractionAgent
from tests.utils_for_test import get_random_state, make_respond, set_16_omni


def test_query_satisfy():
    cmd_records = [
        [
            "sw_card 1 2 3 4 0",
            "choose 1",
            "card 1 0 15",
            "sw_char 2 14 13",
            "end",
            "card 5 0",
            "card 2 3",
            "sw_char 3 15",
            "sw_char 2 14",
            "sw_char 3 13",
            "skill 1 12 11 10",
            "card 4 3",
            "card 0 3 10 9",
            "sw_char 1 9",
            "sw_char 3 8",
            "card 1 0 7 6 5",
            "skill 1 4 3 2",
            "skill 0 2 1 0",
            "end",
            "sw_char 4 15",
            "skill 1 14 13 12",
            "sw_char 2 11",
            "skill 0 10 9 8",
            "sw_char 1 7",
            "sw_char 0 6",
            "end",
            "sw_char 4 15",
            "card 3 0 14 13 12",
            "skill 1 11 10 9",
            "sw_char 1 8",
            "sw_char 0 7",
            "skill 1 6",
            "skill 1 5",
            "sw_char 4 4",
        ],
        [
            "sw_card 0 1 2 3 4",
            "choose 2",
            "card 0 0 15",
            "sw_char 4 13 12",
            "skill 0 11 10 9",
            "end",
            "sw_char 1 15",
            "skill 1 14 13 12",
            "skill 0 11 10 9",
            "skill 0 8 7 6",
            "card 3 1 5",
            "sw_char 3 4",
            "sw_char 0 3",
            "skill 1 2",
            "end",
            "sw_char 1 15",
            "card 4 0 14 13 12",
            "skill 1 11 10 9",
            "sw_char 2 8",
            "sw_char 1 7",
            "skill 0 6 5 4",
            "skill 0 3 4 2",
            "skill 0 1 2 0",
            "card 6 0",
            "end",
            "skill 0 15 14 13",
            "card 3 2",
            "sw_char 2 13",
            "card 2 0 12 11 10",
            "sw_char 4 9",
            "skill 1 8 7 6",
            "skill 0 5 4",
            "skill 0 3 2",
            "card 0 0 1",
            "end",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        "default_version:4.3\ncharacter:Yoimiya@3.8\ncharacter:Keqing@3.3\ncharacter:Dehya@4.1\ncharacter:Shenhe@4.2\ncharacter:Barbara@3.3\nLost Prayer to the Sacred Winds@4.3\nLost Prayer to the Sacred Winds@4.3\nTulaytullah's Remembrance@4.3\nTulaytullah's Remembrance@4.3\nBeacon of the Reed Sea@4.3\nBeacon of the Reed Sea@4.3\nPrimordial Jade Winged-Spear@4.3\nPrimordial Jade Winged-Spear@4.3\nLight of Foliar Incision@4.3\nLight of Foliar Incision@4.3\nFlowing Rings@4.3\nFlowing Rings@4.3\nEchoes of an Offering@4.3\nEchoes of an Offering@4.3\nHeart of Khvarena's Brilliance@4.3\nHeart of Khvarena's Brilliance@4.3\nVourukasha's Glow@4.3\nVourukasha's Glow@4.3\nThe Boar Princess@4.3\nThe Boar Princess@4.3\nFalls and Fortune@4.3\nFalls and Fortune@4.3\n"  # noqa: E501
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    match.config.random_first_player = False
    match.config.initial_hand_size = 5
    match.config.max_hand_size = 99
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR

    # test query_one
    p0c0 = match.player_tables[0].characters[0]
    p1c1 = match.player_tables[1].characters[1]
    p0c4 = match.player_tables[0].characters[4]
    p0 = match.player_tables[0].team_status[0]
    with pytest.raises(ValueError):
        p0c0.query_one(match, "both character")
    with pytest.raises(ValueError):
        p0c0.query_one(match, "our hand 'name=Vourukasha's Glow'")
    with pytest.raises(ValueError):
        p0c0.query_one(match, "both character name=Keqing")
    res = p0c0.query_one(match, "both character weapon 'name=Light of Foliar Incision'")
    assert res is not None and res.name == "Light of Foliar Incision"
    res = p0c0.query_one(match, "both character status 'usage=1'")
    assert res is not None and res.name == "Electro Elemental Infusion"
    with pytest.raises(ValueError):
        res = p0c0.query_one(match, "both character status 'usage=2'")
    with pytest.raises(ValueError):
        res = p0c0.query_one(match, "both active skill 'skill_type=elemental_burst'")
    res = p0c0.query_one(
        match, r"""both active weapon "name=Tulaytullah's Remembrance" """
    )
    assert res is not None and res.name == "Tulaytullah's Remembrance"
    sres: SkillBase = p0c0.query_one(
        match, "our active skill 'skill_type=elemental_burst'"
    )  # type: ignore
    assert (
        sres is not None
        and sres.position.player_idx == 0
        and sres.position.character_idx == 4
        and sres.skill_type == SkillType.ELEMENTAL_BURST
        and sres.name == "Shining Miracle"
    )
    sres = p0c0.query_one(match, "opponent active skill 'skill_type=elemental_burst'")  # type: ignore  # noqa: E501
    assert (
        sres is not None
        and sres.position.player_idx == 1
        and sres.position.character_idx == 4
        and sres.skill_type == SkillType.ELEMENTAL_BURST
        and sres.name == "Shining Miracle"
    )
    res = p0c0.query_one(match, "self status")  # type: ignore
    assert (
        res is not None
        and res.position.player_idx == 0
        and res.position.character_idx == 0
        and res.name == "Niwabi Enshou"
    )
    with pytest.raises(ValueError):
        res = p0c0.query_one(match, "opponent character status")
    res = p0c0.query_one(match, "opponent team_status")
    assert res is None
    res = p0c0.query_one(match, "our team_status")
    assert res is not None and res.name == "Icy Quill"
    with pytest.raises(AssertionError):
        p0c0.query_one(match, "self team_status")
    with pytest.raises(AssertionError):
        p0c0.query_one(match, "both weapon")
    with pytest.raises(AssertionError):
        p0c0.query_one(match, "our weapon")
    res = p0c0.query_one(
        match, "our team_status and opponent team_status and our support"
    )
    assert res is not None and res.name == "Icy Quill"
    with pytest.raises(ValueError):
        p0c0.query_one(match, "our character status and opponent character status")
    res = p0c0.query_one(match, "our character weapon", allow_multiple=True)
    assert res is not None and res.name in [
        "Lost Prayer to the Sacred Winds",
        "Primordial Jade Winged-Spear",
    ]
    with pytest.raises(ValueError):
        p0c0.query_one(match, "character")
    with pytest.raises(ValueError):
        p0c0.query_one(match, "both both both character")
    res = p0c0.query_one(match, "our next status")
    assert res is not None and res.name == "Niwabi Enshou"
    res = p0c0.query_one(match, "opponent prev")
    assert res is not None and res.name == "Shenhe"
    res = p1c1.query_one(match, "our character hp=2")
    assert res is not None and res.name == "Yoimiya"
    res = p1c1.query_one(match, "opponent character status usage=2")
    assert res is not None and res.name == "Niwabi Enshou"

    # test query
    res = p0c0.query(match, "both character")
    assert len(res) == 10
    res = p0c0.query(match, "both character weapon=None")
    assert len(res) == 6
    res = p0c0.query(match, "our character charge=0")
    assert len(res) == 2
    res = p0c0.query(match, "opponent character charge=2")
    assert len(res) == 0
    res = p0c0.query(match, "both next")
    assert len(res) == 2
    for r in res:
        assert r.name == "Yoimiya"
        assert r.position.character_idx == 0
    res = p0c0.query(match, "our summon usage=1")
    assert len(res) == 0
    res = p0c0.query(match, "our summon usage=2")
    assert len(res) == 1 and res[0].name == "Melody Loop"
    res = p0c0.query(match, "both summon usage=2")
    assert len(res) == 2
    for r in res:
        assert r.name == "Melody Loop"
    res = p0c0.query(match, "our hand 'name=Primordial Jade Winged-Spear'")
    assert len(res) == 1 and res[0].name == "Primordial Jade Winged-Spear"
    res = p0c0.query(match, "both hand 'name=Primordial Jade Winged-Spear'")
    assert len(res) == 3
    for r in res:
        assert r.name == "Primordial Jade Winged-Spear"
    res = p0c0.query(match, "both character artifact")
    assert len(res) == 4
    res = p0c0.query(match, "our deck")
    assert len(res) == 9
    res = p1c1.query(match, "our hand 'name=Lost Prayer to the Sacred Winds'")
    assert len(res) == 2
    res = p1c1.query(match, "our hand 'name=Lost Prayer to the Sacred Winds'")
    with pytest.raises(AssertionError):
        p0.query(match, "self")

    # satisfy
    p0c0 = p0c0.position
    p1c1 = p1c1.position
    p0c4 = p0c4.position
    p0 = p0.position
    with pytest.raises(AssertionError):
        p0c0.satisfy("source active=true")
    with pytest.raises(AssertionError):
        p0c0.satisfy("target active=true")
    with pytest.raises(AssertionError):
        p0c0.satisfy("target pidx=1")
    with pytest.raises(AssertionError):
        p0c0.satisfy("source pidx=0 and target pidx=1")
    with pytest.raises(AssertionError):
        p0c0.satisfy("source pidx=0 active=false")
    assert not p0c0.satisfy("source pidx=1")
    assert not p0c0.satisfy("source active=true", match=match)
    assert p0c4.satisfy("source active=true", match=match)
    assert not p0c4.satisfy("source active=true pidx=1", match=match)
    assert p0c4.satisfy("source active=true pidx=0 cidx=4", match=match)
    assert p1c1.satisfy("source pidx=1 cidx=1")
    assert p0c4.satisfy("source pidx=0 cidx=4 and target pidx=1 cidx=1", target=p1c1)
    assert not p0c4.satisfy(
        "source pidx=0 cidx=4 and target pidx=1 cidx=2", target=p1c1
    )
    assert not p0c4.satisfy(
        "source pidx=0 cidx=2 and target pidx=1 cidx=1", target=p1c1
    )
    with pytest.raises(AssertionError):
        p0c0.satisfy("both pidx=same")
    with pytest.raises(ValueError):
        p0c0.satisfy("source active=1", match=match)
    with pytest.raises(ValueError):
        p0c0.satisfy("source sss=1", match=match)
    with pytest.raises(ValueError):
        p0c0.satisfy("source hahaha", match=match)
    with pytest.raises(ValueError):
        p0c0.satisfy("active=1", match=match)
    with pytest.raises(ValueError):
        p0.satisfy("source active=true", match=match)
    assert p1c1.satisfy(f"source id={p1c1.id}")
    assert not p1c1.satisfy(f"source id={p0c4.id}")
    assert p1c1.satisfy("source area=CHARACTER")
    assert p1c1.satisfy("source area=character")
    assert not p1c1.satisfy("source area=team_status")
    assert p0.satisfy("source area=team_status")
    # satisfy both
    with pytest.raises(ValueError):
        assert p0c0.satisfy("both pidx=0", match=match, target=p1c1)
    with pytest.raises(ValueError):
        assert p0c0.satisfy("both id=1", match=match, target=p1c1)
    with pytest.raises(ValueError):
        assert p0c0.satisfy("both id", match=match, target=p1c1)
    with pytest.raises(ValueError):
        assert p0c0.satisfy("both di=same", match=match, target=p1c1)
    assert not p0c0.satisfy("both id=same", match=match, target=p1c1)
    assert p0c0.satisfy("both pidx=diff", match=match, target=p1c1)
    assert p0c0.satisfy("both pidx=same cidx=diff", match=match, target=p0)
    assert not p0c0.satisfy("both pidx=same cidx=same", match=match, target=p0c4)
    assert p0c0.satisfy(
        "both pidx=same cidx=diff and source pidx=0 and target area=character",
        match=match,
        target=p0c4,
    )
    assert not p0c0.satisfy(
        "both pidx=same cidx=diff and source pidx=1 and target area=character",
        match=match,
        target=p0c4,
    )
    assert not p0c0.satisfy(
        "both pidx=same cidx=diff and source pidx=0 and target area=character_status",
        match=match,
        target=p0c4,
    )
    assert not p0c0.satisfy("both pidx=diff cidx=diff", match=match, target=p0)
    assert p0c0.satisfy("both pidx=same cidx=diff id=diff", match=match, target=p0)
    assert not p0c0.satisfy("both pidx=same cidx=diff id=same", match=match, target=p0)
    assert p0c0.satisfy(
        "both pidx=same cidx=diff id=diff area=diff", match=match, target=p0
    )
    assert p0c0.not_satisfy(
        "both pidx=same cidx=diff id=diff area=same", match=match, target=p0
    )
    assert p0c0.not_satisfy(
        "both pidx=same cidx=diff id=diff area=same", match=match, target=p0
    )
    assert p0c0.satisfy(
        "both pidx=same cidx=diff id=diff area=same", match=match, target=p0c4
    )


if __name__ == "__main__":
    test_query_satisfy()



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...summon.base import SwirlChangeSummonBase

from ...event import RoundEndEventArguments

from ...action import (
    Actions,
    ChangeObjectUsageAction,
    MakeDamageAction,
    SwitchCharacterAction,
)
from ...struct import Cost

from ...consts import (
    DamageElementalType,
    DieColor,
    ElementType,
    FactionType,
    WeaponType,
)
from ..character_base import (
    ElementalBurstBase,
    ElementalSkillBase,
    PhysicalNormalAttackBase,
    CharacterBase,
    SkillTalent,
)


# Summons


class Stormeye_3_7(SwirlChangeSummonBase):
    name: Literal["Stormeye"] = "Stormeye"
    version: Literal["3.7"] = "3.7"
    usage: int = 2
    max_usage: int = 2
    damage: int = 2

    def event_handler_ROUND_END(
        self, event: RoundEndEventArguments, match: Any
    ) -> List[MakeDamageAction | ChangeObjectUsageAction | SwitchCharacterAction]:
        """
        Need to change enemy active
        """
        ret: List[
            MakeDamageAction | ChangeObjectUsageAction | SwitchCharacterAction
        ] = []
        ret += super().event_handler_ROUND_END(event, match)
        our_active = match.player_tables[self.position.player_idx].active_character_idx
        target_characters = match.player_tables[1 - self.position.player_idx].characters
        target_idx = our_active
        if target_characters[target_idx].is_defeated:
            # need to choose another character
            for idx, character in enumerate(target_characters):
                if not character.is_defeated:
                    target_idx = idx
                    break
            else:
                raise AssertionError("No character alive")
        if (
            target_idx
            != match.player_tables[1 - self.position.player_idx].active_character_idx
        ):
            ret.append(
                SwitchCharacterAction(
                    player_idx=1 - self.position.player_idx,
                    character_idx=target_idx,
                )
            )
        return ret


# Skills


class SkywardSonnet(ElementalSkillBase):
    name: Literal["Skyward Sonnet"] = "Skyward Sonnet"
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.ANEMO
    cost: Cost = Cost(elemental_dice_color=DieColor.ANEMO, elemental_dice_number=3)

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Attack and create object
        """
        character = match.player_tables[self.position.player_idx].characters[
            self.position.character_idx
        ]
        assert character.name == "Venti"
        talent_activated = False
        if self.is_talent_equipped(match):
            talent_activated = True
        return super().get_actions(match) + [
            self.create_team_status("Stormzone", {"talent_activated": talent_activated})
        ]


class WindsGrandOde(ElementalBurstBase):
    name: Literal["Wind's Grand Ode"] = "Wind's Grand Ode"
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.ANEMO
    cost: Cost = Cost(
        elemental_dice_color=DieColor.ANEMO, elemental_dice_number=3, charge=2
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Attack and create object
        """
        # create summon first, so it can change element immediately.
        return super().get_actions(match) + [self.create_summon("Stormeye")]


# Talents


class EmbraceOfWinds_3_7(SkillTalent):
    name: Literal["Embrace of Winds"]
    version: Literal["3.7"] = "3.7"
    character_name: Literal["Venti"] = "Venti"
    cost: Cost = Cost(elemental_dice_color=DieColor.ANEMO, elemental_dice_number=3)
    skill: Literal["Skyward Sonnet"] = "Skyward Sonnet"


# character base


class Venti_3_7(CharacterBase):
    name: Literal["Venti"]
    version: Literal["3.7"] = "3.7"
    element: ElementType = ElementType.ANEMO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[PhysicalNormalAttackBase | SkywardSonnet | WindsGrandOde] = []
    faction: List[FactionType] = [FactionType.MONDSTADT]
    weapon_type: WeaponType = WeaponType.BOW

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name="Divine Marksmanship",
                cost=PhysicalNormalAttackBase.get_cost(self.element),
            ),
            SkywardSonnet(),
            WindsGrandOde(),
        ]


register_class(Venti_3_7 | EmbraceOfWinds_3_7 | Stormeye_3_7)



from typing import Dict, List, Literal
from pydantic import PrivateAttr

from ....event_handler import SystemEventHandlerBase
from ....object_base import CreateSystemEventHandlerObject
from .....utils.class_registry import register_class
from ....struct import Cost
from ....action import ChangeObjectUsageAction, DrawCardAction, RemoveObjectAction
from ....match import Match
from ....event import (
    MoveObjectEventArguments,
    ReceiveDamageEventArguments,
    RoundEndEventArguments,
)
from ....consts import DamageElementalType, DamageType, IconType
from ....card.support.companions import CompanionBase
from .....utils.desc_registry import DescDictType


class SilverAndMelusEventHandler_4_4(SystemEventHandlerBase):
    """
    It will be generated when Silver and Melus in deck, and record elemental damage
    types for both player, and when Silver and Melus is played into area or elemental
    damage is made, it will update its counter and also update all existing Silver and
    Melus usage.
    """

    name: Literal["Silver and Melus"] = "Silver and Melus"
    version: Literal["4.4"] = "4.4"
    etypes: List[List[DamageElementalType]] = [[], []]
    _max_usage: int = PrivateAttr(4)

    def _update_usage(self, match: Match) -> List[ChangeObjectUsageAction]:
        """
        check all, if their usage not same as self, update it.
        """
        actions: List[ChangeObjectUsageAction] = []
        for etypes, table in zip(self.etypes, match.player_tables):
            counter = min(len(etypes), self._max_usage)
            for support in table.supports:
                if support.name == self.name and support.usage != counter:
                    actions.append(
                        ChangeObjectUsageAction(
                            object_position=support.position,
                            change_usage=counter - support.usage,
                        )
                    )
        return actions

    def event_handler_MOVE_OBJECT(
        self, event: MoveObjectEventArguments, match: Match
    ) -> List[ChangeObjectUsageAction]:
        if event.object_name != self.name:
            # not Silver and Melus
            return []
        return self._update_usage(match)

    def event_handler_RECEIVE_DAMAGE(
        self, event: ReceiveDamageEventArguments, match: Match
    ) -> List[ChangeObjectUsageAction]:
        pidx = event.final_damage.target_position.player_idx
        if event.final_damage.damage_type != DamageType.DAMAGE or (
            event.final_damage.damage_elemental_type
            in [
                DamageElementalType.PIERCING,
                DamageElementalType.PHYSICAL,
                DamageElementalType.HEAL,
            ]
        ):
            # not damage or damage type not elemental
            return []
        # record damage type
        damage_element = event.final_damage.damage_elemental_type
        # damage is too frequent. If not new damage type, do not check usage.
        changed = False
        if damage_element not in self.etypes[1 - pidx]:
            self.etypes[1 - pidx].append(damage_element)
            changed = True
        if not changed:
            return []
        return self._update_usage(match)


class SilverAndMelus_4_4(CreateSystemEventHandlerObject, CompanionBase):
    name: Literal["Silver and Melus"]
    version: Literal["4.4"] = "4.4"
    usage: int = 0
    _max_usage: int = PrivateAttr(4)
    recorded_damage_type: List[DamageElementalType] = []
    icon_type: Literal[IconType.TIMESTATE] = IconType.TIMESTATE
    cost: Cost = Cost(same_dice_number=1)

    handler_name: Literal["Silver and Melus"] = "Silver and Melus"

    def event_handler_ROUND_END(
        self,
        event: RoundEndEventArguments,
        match: Match,
    ) -> List[DrawCardAction | RemoveObjectAction]:
        """
        If usage above 3, remove self and draw cards.
        """
        if self.usage < 3:
            return []
        return [
            DrawCardAction(
                player_idx=self.position.player_idx,
                number=self.usage,
                draw_if_filtered_not_enough=True,
            ),
            RemoveObjectAction(object_position=self.position),
        ]


desc: Dict[str, DescDictType] = {
    "SUPPORT/Silver and Melus": {
        "names": {"en-US": "Silver and Melus", "zh-CN": "西尔弗和迈勒斯"},
        "descs": {
            "4.4": {
                "en-US": "This card will record the number of types of Elemental DMG opposing characters have taken during this match as points of Attendant's Attentiveness. (Max 4 points)\nEnd Phase: If you have at least 3 Attendant's Attentiveness points, discard this card and draw cards equal to the amount of Attendant's Attentiveness points.",  # noqa: E501
                "zh-CN": "此牌会记录本场对局中敌方角色受到过的元素伤害种类数，称为「侍从的周到」。（最多4点）\n结束阶段：如果「侍从的周到」至少为3，则弃置此牌，然后抓「侍从的周到」点数的牌。",  # noqa: E501
            }
        },
        "image_path": "cardface/Assist_NPC_Silver.png",  # noqa: E501
        "id": 322023,
    },
    "SYSTEM/Silver and Melus": {
        "names": {"en-US": "Silver and Melus", "zh-CN": "西尔弗和迈勒斯"},
        "descs": {"4.4": {}},
    },
}


register_class(SilverAndMelus_4_4 | SilverAndMelusEventHandler_4_4, desc)



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...modifiable_values import CostValue
from ...event import CharacterReviveEventArguments, RoundPrepareEventArguments

from ...action import Actions
from ...struct import Cost

from ...consts import (
    DamageElementalType,
    DieColor,
    ElementType,
    FactionType,
    WeaponType,
)
from ..character_base import (
    ElementalBurstBase,
    ElementalSkillBase,
    PhysicalNormalAttackBase,
    CharacterBase,
    SkillTalent,
)
# Skills


class SearingOnslaught(ElementalSkillBase):
    name: Literal["Searing Onslaught"] = "Searing Onslaught"
    damage_type: DamageElementalType = DamageElementalType.PYRO
    cost: Cost = Cost(elemental_dice_color=DieColor.PYRO, elemental_dice_number=3)

    counter: int = 0

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Any
    ) -> List[Actions]:
        """
        When in round prepare, reset counter
        """
        self.counter = 0
        return []

    def event_handler_CHARACTER_REVIVE(
        self, event: CharacterReviveEventArguments, match: Any
    ) -> List[Actions]:
        """
        When self is revived, clear counter
        """
        if (
            event.action.player_idx == self.position.player_idx
            and event.action.character_idx == self.position.character_idx
        ):
            self.counter = 0
        return []

    def value_modifier_COST(
        self, value: CostValue, match: Any, mode: Literal["REAL", "TEST"]
    ) -> CostValue:
        """
        If self equipped talent, reduce the second use cost by 1
        """
        if self.counter != 1:
            # not second use
            return value
        if not self.is_talent_equipped(match):
            # no talent
            return value
        if not self.position.check_position_valid(
            value.position,
            match,
            player_idx_same=True,
            character_idx_same=True,
            area_same=True,
            id_same=True,
        ):
            # not this skill
            return value
        # decrease cost
        value.cost.decrease_cost(DieColor.PYRO)
        return value

    def get_actions(self, match: Any) -> List[Actions]:
        """
        if it is the third time to use, increase damage by 2
        """
        self.counter += 1
        if self.counter == 3:
            self.damage = 5
        ret = super().get_actions(match)
        self.damage = 3
        return ret


class Dawn(ElementalBurstBase):
    name: Literal["Dawn"] = "Dawn"
    damage: int = 8
    damage_type: DamageElementalType = DamageElementalType.PYRO
    cost: Cost = Cost(
        elemental_dice_color=DieColor.PYRO, elemental_dice_number=4, charge=3
    )

    def get_actions(self, match: Any) -> List[Actions]:
        return super().get_actions(match) + [
            self.create_character_status("Pyro Elemental Infusion", {"mark": "Diluc"})
        ]


# Talents


class FlowingFlame_3_3(SkillTalent):
    name: Literal["Flowing Flame"]
    version: Literal["3.3"] = "3.3"
    character_name: Literal["Diluc"] = "Diluc"
    cost: Cost = Cost(
        elemental_dice_color=DieColor.PYRO,
        elemental_dice_number=3,
    )
    skill: Literal["Searing Onslaught"] = "Searing Onslaught"


# character base


class Diluc_3_3(CharacterBase):
    name: Literal["Diluc"]  # Do not set default value for character name
    version: Literal["3.3"] = "3.3"
    element: ElementType = ElementType.PYRO
    max_hp: int = 10
    max_charge: int = 3
    skills: List[PhysicalNormalAttackBase | SearingOnslaught | Dawn] = []
    faction: List[FactionType] = [FactionType.MONDSTADT]
    weapon_type: WeaponType = WeaponType.CLAYMORE

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name="Tempered Sword",
                cost=PhysicalNormalAttackBase.get_cost(self.element),
            ),
            SearingOnslaught(),
            Dawn(),
        ]


register_class(Diluc_3_3 | FlowingFlame_3_3)



from typing import Any, Dict, List, Literal

from pydantic import PrivateAttr

from lpsim.server.consts import (
    CostLabels,
    IconType,
    ObjectPositionType,
    PlayerActionLabels,
)
from lpsim.server.match import Match
from lpsim.server.status.team_status.base import (
    ShieldTeamStatus,
    SwitchActionTeamStatus,
    UsageTeamStatus,
)
from lpsim.utils.class_registry import register_class
from lpsim.utils.desc_registry import DescDictType
from lpsim.server.modifiable_values import CostValue, DamageValue
from lpsim.server.event import ActionEndEventArguments, RoundPrepareEventArguments
from lpsim.server.action import (
    Actions,
    CreateObjectAction,
    DrawCardAction,
    MakeDamageAction,
)
from lpsim.server.struct import Cost, ObjectPosition
from lpsim.consts import (
    DamageElementalType,
    DamageType,
    DieColor,
    ElementType,
    FactionType,
    WeaponType,
)
from lpsim.server.character.character_base import (
    ElementalBurstBase,
    ElementalSkillBase,
    PhysicalNormalAttackBase,
    CharacterBase,
    SkillTalent,
)


# Team status.


class UrgentNekoParcel_4_5(SwitchActionTeamStatus):
    name: Literal["Urgent Neko Parcel"] = "Urgent Neko Parcel"
    version: Literal["4.5"] = "4.5"
    usage: int = 1
    max_usage: int = 1
    icon_type: IconType = IconType.OTHERS

    def _action(self, match: Match) -> List[MakeDamageAction | DrawCardAction]:
        self.usage -= 1
        target = self.query_one(match, "opponent active")
        assert target is not None
        return [
            MakeDamageAction(
                damage_value_list=[
                    DamageValue(
                        position=self.position,
                        damage_type=DamageType.DAMAGE,
                        target_position=target.position,
                        damage=1,
                        damage_elemental_type=DamageElementalType.DENDRO,
                        cost=Cost(),
                    )
                ]
            ),
            DrawCardAction(
                player_idx=self.position.player_idx,
                number=1,
                draw_if_filtered_not_enough=True,
            ),
        ]


class ShieldOfSafeTransport_4_5(ShieldTeamStatus):
    name: Literal["Shield of Safe Transport"] = "Shield of Safe Transport"
    version: Literal["4.5"] = "4.5"
    usage: int = 2
    max_usage: int = 2


class CatGrassCardamom_4_5(UsageTeamStatus):
    name: Literal["Cat Grass Cardamom"] = "Cat Grass Cardamom"
    version: Literal["4.5"] = "4.5"
    usage: int = 2
    max_usage: int = 2
    counter: int = 0
    icon_type: IconType = IconType.OTHERS

    def event_handler_ACTION_END(
        self, event: ActionEndEventArguments, match: Match
    ) -> List[MakeDamageAction]:
        if (
            event.action.position.player_idx != self.position.player_idx
            or event.action.action_label & PlayerActionLabels.CARD.value == 0
        ):
            # not self player, or not use card
            return []
        self.counter += 1
        if self.counter == 2:
            target = self.query_one(match, "our active")
            self.usage -= 1
            self.counter = 0
            assert target is not None
            return [
                MakeDamageAction(
                    damage_value_list=[
                        DamageValue(
                            position=self.position,
                            damage_type=DamageType.DAMAGE,
                            target_position=target.position,
                            damage=1,
                            damage_elemental_type=DamageElementalType.DENDRO,
                            cost=Cost(),
                        )
                    ]
                )
            ]
        return []


# Skills


class MeowteorKick(ElementalSkillBase):
    name: Literal["Meow-teor Kick"] = "Meow-teor Kick"
    damage: int = 0
    damage_type: DamageElementalType = DamageElementalType.DENDRO
    cost: Cost = Cost(elemental_dice_color=DieColor.DENDRO, elemental_dice_number=3)

    def get_actions(self, match: Match) -> List[Actions]:
        return [
            self.create_team_status("Urgent Neko Parcel"),
            self.create_team_status("Shield of Safe Transport"),
            self.charge_self(1),
        ]


class SecretArtSurpriseDispatch(ElementalBurstBase):
    name: Literal["Secret Art: Surprise Dispatch"] = "Secret Art: Surprise Dispatch"
    damage: int = 4
    damage_type: DamageElementalType = DamageElementalType.DENDRO
    cost: Cost = Cost(
        elemental_dice_color=DieColor.DENDRO, elemental_dice_number=3, charge=2
    )

    def get_actions(self, match: Match) -> List[Actions]:
        """
        Attack and create opposite
        """
        return super().get_actions(match) + [
            CreateObjectAction(
                object_name="Cat Grass Cardamom",
                object_position=ObjectPosition(
                    player_idx=1 - self.position.player_idx,
                    area=ObjectPositionType.TEAM_STATUS,
                    id=0,
                ),
                object_arguments={},
            )
        ]


# Talents


class CountlessSightsToSee_4_5(SkillTalent):
    name: Literal["Countless Sights to See"]
    version: Literal["4.5"] = "4.5"
    character_name: Literal["Kirara"] = "Kirara"
    cost: Cost = Cost(elemental_dice_color=DieColor.DENDRO, elemental_dice_number=3)
    skill: Literal["Meow-teor Kick"] = "Meow-teor Kick"

    usage: int = 1
    _usage_per_round: int = PrivateAttr(1)

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Match
    ) -> List[Actions]:
        self.usage = self._usage_per_round
        return []

    def get_actions(self, target: ObjectPosition | None, match: Any) -> List[Actions]:
        self.usage = self._usage_per_round
        return super().get_actions(target, match)

    def value_modifier_COST(
        self, value: CostValue, match: Match, mode: Literal["TEST", "REAL"]
    ) -> CostValue:
        if self.usage == 0:
            return value
        if value.cost.label & CostLabels.SWITCH_CHARACTER.value == 0:
            # not switch character
            return value
        if self.position.not_satisfy(
            "both pidx=same cidx=same and source active=true",
            value.position,
            match=match,
        ):
            # not cost from self character to others
            return value
        value.cost.decrease_cost(None)
        if mode == "REAL":
            self.usage -= 1
        return value


# character base


# character class name should contain its version.
class Kirara_4_5(CharacterBase):
    name: Literal["Kirara"]
    version: Literal["4.5"] = "4.5"
    element: ElementType = ElementType.DENDRO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[
        PhysicalNormalAttackBase | MeowteorKick | SecretArtSurpriseDispatch
    ] = []
    faction: List[FactionType] = [FactionType.INAZUMA]
    weapon_type: WeaponType = WeaponType.SWORD

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name="Boxcutter",
                cost=PhysicalNormalAttackBase.get_cost(self.element),
            ),
            MeowteorKick(),
            SecretArtSurpriseDispatch(),
        ]


# define descriptions of newly defined classes. Note key of skills and talents
# have character names. For balance changes, only descs are needed to define.
character_descs: Dict[str, DescDictType] = {
    "CHARACTER/Kirara": {
        "names": {"en-US": "Kirara", "zh-CN": "绮良良"},
        "descs": {"4.5": {"en-US": "", "zh-CN": ""}},
        "image_path": "cardface/Char_Avatar_Momoka.png",  # noqa: E501
        "id": 1707,
    },
    "SKILL_Kirara_NORMAL_ATTACK/Boxcutter": {
        "names": {"en-US": "Boxcutter", "zh-CN": "箱纸切削术"},
        "descs": {
            "4.5": {"en-US": "Deals 2 Physical DMG.", "zh-CN": "造成2点物理伤害。"}
        },
    },
    "SKILL_Kirara_ELEMENTAL_SKILL/Meow-teor Kick": {
        "names": {"en-US": "Meow-teor Kick", "zh-CN": "呜喵町飞足"},
        "descs": {
            "4.5": {
                "en-US": "Creates 1 Urgent Neko Parcel and Shield of Safe Transport.",
                "zh-CN": "生成猫箱急件和安全运输护盾。",
            }
        },
    },
    "TEAM_STATUS/Urgent Neko Parcel": {
        "names": {"en-US": "Urgent Neko Parcel", "zh-CN": "猫箱急件"},
        "descs": {
            "4.5": {
                "en-US": "When Kirara is your active character, after you switch characters: Deal 1 Dendro DMG and draw 1 card.\nUsage(s): 1 (Can stack. Max 2 stacks.)",  # noqa: E501
                "zh-CN": "绮良良为出战角色时，我方切换角色后：造成1点草元素伤害，抓1张牌。\n可用次数：1（可叠加，最多叠加到2次）",  # noqa: E501
            }
        },
        "image_path": "status/Momoka_S.png",
    },
    "TEAM_STATUS/Shield of Safe Transport": {
        "names": {"en-US": "Shield of Safe Transport", "zh-CN": "安全运输护盾"},
        "descs": {
            "4.5": {
                "en-US": "Shield\nGrants 2 Shield points to your active character.",
                "zh-CN": "护盾\n为我方出战角色提供2点护盾。",
            }
        },
    },
    "SKILL_Kirara_ELEMENTAL_BURST/Secret Art: Surprise Dispatch": {
        "names": {"en-US": "Secret Art: Surprise Dispatch", "zh-CN": "秘法·惊喜特派"},
        "descs": {
            "4.5": {
                "en-US": "Deals 4 Dendro DMG, creates 1 Cat Grass Cardamom on the opponent's side of the field.",  # noqa: E501
                "zh-CN": "造成4点草元素伤害，在敌方场上生成猫草豆蔻。",
            }
        },
    },
    "TEAM_STATUS/Cat Grass Cardamom": {
        "names": {"en-US": "Cat Grass Cardamom", "zh-CN": "猫草豆蔻"},
        "descs": {
            "4.5": {
                "en-US": "After the side of the field that this card is on has played 2 Action Cards: Deal 1 Dendro DMG to the active character on that side.\nUsage(s): 2",  # noqa: E501
                "zh-CN": "所在阵营打出2张行动牌后：对所在阵营的出战角色造成1点草元素伤害。\n可用次数：2",  # noqa: E501
            }
        },
        "image_path": "status/Debuff_Momoka_E.png",
    },
    "TALENT_Kirara/Countless Sights to See": {
        "names": {"en-US": "Countless Sights to See", "zh-CN": "沿途百景会心"},
        "descs": {
            "4.5": {
                "en-US": "Combat Action: When your active character is Kirara, equip this card.\nAfter Kirara equips this card, immediately use Meow-teor Kick once.\nWhen your Kirara who has this card equipped is the active character, spend 1 less Elemental Die when switching characters. (Once per Round)\n(You must have Kirara in your deck to add this card to your deck.)",  # noqa: E501
                "zh-CN": "战斗行动：我方出战角色为绮良良时，装备此牌。\n绮良良装备此牌后，立刻使用一次呜喵町飞足。\n装备有此牌的绮良良为出战角色，我方进行「切换角色」行动时：少花费1个元素骰。（每回合1次）\n（牌组中包含绮良良，才能加入牌组）",  # noqa: E501
            }
        },
        "image_path": "cardface/Modify_Talent_Momoka.png",  # noqa: E501
        "id": 217071,
    },
}


register_class(
    Kirara_4_5
    | UrgentNekoParcel_4_5
    | ShieldOfSafeTransport_4_5
    | CatGrassCardamom_4_5
    | CountlessSightsToSee_4_5,
    character_descs,
)



from typing import Any, List, Literal
from lpsim.server.action import (
    ConsumeArcaneLegendAction,
    CreateObjectAction,
    RemoveObjectAction,
)

from lpsim.server.card.event.arcane_legend import ArcaneLegendBase
from lpsim.server.consts import IconType, ObjectPositionType
from lpsim.server.event import RoundPrepareEventArguments
from lpsim.server.match import Match
from lpsim.server.status.character_status.base import DefendCharacterStatus
from lpsim.server.struct import Cost, ObjectPosition
from lpsim.utils.class_registry import register_class
from lpsim.utils.desc_registry import DescDictType


class DayOfResistanceMomentOfShatteredDreamsStatus_4_5(DefendCharacterStatus):
    name: Literal[
        "Day of Resistance: Moment of Shattered Dreams"
    ] = "Day of Resistance: Moment of Shattered Dreams"
    version: Literal["4.5"] = "4.5"
    usage: int = 4
    max_usage: int = 4
    min_damage_to_trigger: int = 1
    max_in_one_time: int = 1
    icon_type: IconType = IconType.BARRIER

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Match
    ) -> List[RemoveObjectAction]:
        return [
            RemoveObjectAction(
                object_position=self.position,
            )
        ]


class DayOfResistanceMomentOfShatteredDreams_4_5(ArcaneLegendBase):
    name: Literal[
        "Day of Resistance: Moment of Shattered Dreams"
    ] = "Day of Resistance: Moment of Shattered Dreams"
    version: Literal["4.5"] = "4.5"
    cost: Cost = Cost(arcane_legend=True)

    def get_targets(self, match: Match) -> List[ObjectPosition]:
        target = self.query(match, "our character is_alive=True")
        target_pos = [x.position for x in target]
        return target_pos

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[ConsumeArcaneLegendAction | CreateObjectAction]:
        assert target is not None
        ret: List[ConsumeArcaneLegendAction | CreateObjectAction] = list(
            super().get_actions(target, match)
        )
        ret.append(
            CreateObjectAction(
                object_name=self.name,
                object_position=target.set_area(ObjectPositionType.CHARACTER_STATUS),
                object_arguments={},
            )
        )
        return ret


desc: dict[str, DescDictType] = {
    "CHARACTER_STATUS/Day of Resistance: Moment of Shattered Dreams": {
        "names": {
            "en-US": "Day of Resistance: Moment of Shattered Dreams",
            "zh-CN": "抗争之日·碎梦之时",
        },
        "descs": {
            "4.5": {
                "en-US": "During this Round, this character takes -1 DMG. (Max 4 times)",  # noqa: E501
                "zh-CN": "本回合中，角色受到的伤害-1。（最多生效4次）",
            }
        },
    },
    "ARCANE/Day of Resistance: Moment of Shattered Dreams": {
        "names": {
            "en-US": "Day of Resistance: Moment of Shattered Dreams",
            "zh-CN": "抗争之日·碎梦之时",
        },
        "descs": {
            "4.5": {
                "en-US": 'During this Round, your targeted character takes -1 DMG. (Max 4 times)\n(Only one "Arcane Legend" card can be played per match. This card will be in your starting hand.)',  # noqa: E501
                "zh-CN": "本回合中，目标我方角色受到的伤害-1。（最多生效4次）\n（整局游戏只能打出一张「秘传」卡牌；这张牌一定在你的起始手牌中）",  # noqa: E501
            }
        },
        "image_path": "cardface/Event_Event_Fankang.png",  # noqa: E501
        "id": 330007,
    },
}


register_class(
    DayOfResistanceMomentOfShatteredDreams_4_5
    | DayOfResistanceMomentOfShatteredDreamsStatus_4_5,
    desc,
)
